#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from nfs4.x on Fri Jan 20 14:19:56 2017
"""
NFSv4 constants module
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "4.2"

# Enum nfs_bool
FALSE = 0
TRUE  = 1

nfs_bool = {
    0 : "FALSE",
    1 : "TRUE",
}

# Sizes
NFS4_FHSIZE         = 128
NFS4_VERIFIER_SIZE  = 8
NFS4_OPAQUE_LIMIT   = 1024
NFS4_OTHER_SIZE     = 12
# Sizes new to NFSv4.1
NFS4_SESSIONID_SIZE = 16
NFS4_DEVICEID4_SIZE = 16
NFS4_INT64_MAX      = 0x7fffffffffffffff
NFS4_UINT64_MAX     = 0xffffffffffffffff
NFS4_INT32_MAX      = 0x7fffffff
NFS4_UINT32_MAX     = 0xffffffff

# Enum nfs_ftype4
NF4REG       = 1  # Regular File
NF4DIR       = 2  # Directory
NF4BLK       = 3  # Special File - block device
NF4CHR       = 4  # Special File - character device
NF4LNK       = 5  # Symbolic Link
NF4SOCK      = 6  # Special File - socket
NF4FIFO      = 7  # Special File - fifo
NF4ATTRDIR   = 8  # Attribute Directory
NF4NAMEDATTR = 9  # Named Attribute

nfs_ftype4 = {
    1 : "NF4REG",
    2 : "NF4DIR",
    3 : "NF4BLK",
    4 : "NF4CHR",
    5 : "NF4LNK",
    6 : "NF4SOCK",
    7 : "NF4FIFO",
    8 : "NF4ATTRDIR",
    9 : "NF4NAMEDATTR",
}

# Enum nfsstat4
NFS4_OK                           = 0      # everything is okay
NFS4ERR_PERM                      = 1      # caller not privileged
NFS4ERR_NOENT                     = 2      # no such file/directory
NFS4ERR_IO                        = 5      # hard I/O error
NFS4ERR_NXIO                      = 6      # no such device
NFS4ERR_ACCESS                    = 13     # access denied
NFS4ERR_EXIST                     = 17     # file already exists
NFS4ERR_XDEV                      = 18     # different filesystems
# Unused/reserved                   19
NFS4ERR_NOTDIR                    = 20     # should be a directory
NFS4ERR_ISDIR                     = 21     # should not be directory
NFS4ERR_INVAL                     = 22     # invalid argument
NFS4ERR_FBIG                      = 27     # file exceeds server max
NFS4ERR_NOSPC                     = 28     # no space on filesystem
NFS4ERR_ROFS                      = 30     # read-only filesystem
NFS4ERR_MLINK                     = 31     # too many hard links
NFS4ERR_NAMETOOLONG               = 63     # name exceeds server max
NFS4ERR_NOTEMPTY                  = 66     # directory not empty
NFS4ERR_DQUOT                     = 69     # hard quota limit reached
NFS4ERR_STALE                     = 70     # file no longer exists
NFS4ERR_BADHANDLE                 = 10001  # Illegal filehandle
NFS4ERR_BAD_COOKIE                = 10003  # READDIR cookie is stale
NFS4ERR_NOTSUPP                   = 10004  # operation not supported
NFS4ERR_TOOSMALL                  = 10005  # response limit exceeded
NFS4ERR_SERVERFAULT               = 10006  # undefined server error
NFS4ERR_BADTYPE                   = 10007  # type invalid for CREATE
NFS4ERR_DELAY                     = 10008  # file "busy" - retry
NFS4ERR_SAME                      = 10009  # nverify says attrs same
NFS4ERR_DENIED                    = 10010  # lock unavailable
NFS4ERR_EXPIRED                   = 10011  # lock lease expired
NFS4ERR_LOCKED                    = 10012  # I/O failed due to lock
NFS4ERR_GRACE                     = 10013  # in grace period
NFS4ERR_FHEXPIRED                 = 10014  # filehandle expired
NFS4ERR_SHARE_DENIED              = 10015  # share reserve denied
NFS4ERR_WRONGSEC                  = 10016  # wrong security flavor
NFS4ERR_CLID_INUSE                = 10017  # clientid in use
# NFS4ERR_RESOURCE is not a valid error in NFSv4.1
NFS4ERR_RESOURCE                  = 10018  # resource exhaustion
NFS4ERR_MOVED                     = 10019  # filesystem relocated
NFS4ERR_NOFILEHANDLE              = 10020  # current FH is not set
NFS4ERR_MINOR_VERS_MISMATCH       = 10021  # minor vers not supp
NFS4ERR_STALE_CLIENTID            = 10022  # server has rebooted
NFS4ERR_STALE_STATEID             = 10023  # server has rebooted
NFS4ERR_OLD_STATEID               = 10024  # state is out of sync
NFS4ERR_BAD_STATEID               = 10025  # incorrect stateid
NFS4ERR_BAD_SEQID                 = 10026  # request is out of seq.
NFS4ERR_NOT_SAME                  = 10027  # verify - attrs not same
NFS4ERR_LOCK_RANGE                = 10028  # overlapping lock range
NFS4ERR_SYMLINK                   = 10029  # should be file/directory
NFS4ERR_RESTOREFH                 = 10030  # no saved filehandle
NFS4ERR_LEASE_MOVED               = 10031  # some filesystem moved
NFS4ERR_ATTRNOTSUPP               = 10032  # recommended attr not sup
NFS4ERR_NO_GRACE                  = 10033  # reclaim outside of grace
NFS4ERR_RECLAIM_BAD               = 10034  # reclaim error at server
NFS4ERR_RECLAIM_CONFLICT          = 10035  # conflict on reclaim
NFS4ERR_BADXDR                    = 10036  # XDR decode failed
NFS4ERR_LOCKS_HELD                = 10037  # file locks held at CLOSE
NFS4ERR_OPENMODE                  = 10038  # conflict in OPEN and I/O
NFS4ERR_BADOWNER                  = 10039  # owner translation bad
NFS4ERR_BADCHAR                   = 10040  # utf-8 char not supported
NFS4ERR_BADNAME                   = 10041  # name not supported
NFS4ERR_BAD_RANGE                 = 10042  # lock range not supported
NFS4ERR_LOCK_NOTSUPP              = 10043  # no atomic up/downgrade
NFS4ERR_OP_ILLEGAL                = 10044  # undefined operation
NFS4ERR_DEADLOCK                  = 10045  # file locking deadlock
NFS4ERR_FILE_OPEN                 = 10046  # open file blocks op.
NFS4ERR_ADMIN_REVOKED             = 10047  # lockowner state revoked
NFS4ERR_CB_PATH_DOWN              = 10048  # callback path down

# NFSv4.1 errors start here
NFS4ERR_BADIOMODE                 = 10049
NFS4ERR_BADLAYOUT                 = 10050
NFS4ERR_BAD_SESSION_DIGEST        = 10051
NFS4ERR_BADSESSION                = 10052
NFS4ERR_BADSLOT                   = 10053
NFS4ERR_COMPLETE_ALREADY          = 10054
NFS4ERR_CONN_NOT_BOUND_TO_SESSION = 10055
NFS4ERR_DELEG_ALREADY_WANTED      = 10056
NFS4ERR_BACK_CHAN_BUSY            = 10057  # backchan reqs outstanding
NFS4ERR_LAYOUTTRYLATER            = 10058
NFS4ERR_LAYOUTUNAVAILABLE         = 10059
NFS4ERR_NOMATCHING_LAYOUT         = 10060
NFS4ERR_RECALLCONFLICT            = 10061
NFS4ERR_UNKNOWN_LAYOUTTYPE        = 10062
NFS4ERR_SEQ_MISORDERED            = 10063  # unexpected seq.id in req
NFS4ERR_SEQUENCE_POS              = 10064  # [CB_]SEQ. op not 1st op
NFS4ERR_REQ_TOO_BIG               = 10065  # request too big
NFS4ERR_REP_TOO_BIG               = 10066  # reply too big
NFS4ERR_REP_TOO_BIG_TO_CACHE      = 10067  # rep. not all cached
NFS4ERR_RETRY_UNCACHED_REP        = 10068  # retry & rep. uncached
NFS4ERR_UNSAFE_COMPOUND           = 10069  # retry/recovery too hard
NFS4ERR_TOO_MANY_OPS              = 10070  # too many ops in [CB_]COMP
NFS4ERR_OP_NOT_IN_SESSION         = 10071  # op needs [CB_]SEQ. op
NFS4ERR_HASH_ALG_UNSUPP           = 10072  # hash alg. not supp.
# Unused/reserved                   10073
NFS4ERR_CLIENTID_BUSY             = 10074  # clientid has state
NFS4ERR_PNFS_IO_HOLE              = 10075  # IO to _SPARSE file hole
NFS4ERR_SEQ_FALSE_RETRY           = 10076  # Retry != original req.
NFS4ERR_BAD_HIGH_SLOT             = 10077  # req has bad highest_slot
NFS4ERR_DEADSESSION               = 10078  # new req sent to dead sess
NFS4ERR_ENCR_ALG_UNSUPP           = 10079  # encr alg. not supp.
NFS4ERR_PNFS_NO_LAYOUT            = 10080  # I/O without a layout
NFS4ERR_NOT_ONLY_OP               = 10081  # addl ops not allowed
NFS4ERR_WRONG_CRED                = 10082  # op done by wrong cred
NFS4ERR_WRONG_TYPE                = 10083  # op on wrong type object
NFS4ERR_DIRDELEG_UNAVAIL          = 10084  # delegation not avail.
NFS4ERR_REJECT_DELEG              = 10085  # cb rejected delegation
NFS4ERR_RETURNCONFLICT            = 10086  # layout get before return
NFS4ERR_DELEG_REVOKED             = 10087  # no return-state revoked

# NFSv4.2 errors start here
NFS4ERR_PARTNER_NOTSUPP           = 10088  # s2s not supported
NFS4ERR_PARTNER_NO_AUTH           = 10089  # s2s not authorized
NFS4ERR_UNION_NOTSUPP             = 10090  # Arm of union not supp
NFS4ERR_OFFLOAD_DENIED            = 10091  # dest not allowing copy
NFS4ERR_WRONG_LFS                 = 10092  # LFS not supported
NFS4ERR_BADLABEL                  = 10093  # incorrect label
NFS4ERR_OFFLOAD_NO_REQS           = 10094  # dest not meeting reqs

nfsstat4 = {
        0 : "NFS4_OK",
        1 : "NFS4ERR_PERM",
        2 : "NFS4ERR_NOENT",
        5 : "NFS4ERR_IO",
        6 : "NFS4ERR_NXIO",
       13 : "NFS4ERR_ACCESS",
       17 : "NFS4ERR_EXIST",
       18 : "NFS4ERR_XDEV",
       20 : "NFS4ERR_NOTDIR",
       21 : "NFS4ERR_ISDIR",
       22 : "NFS4ERR_INVAL",
       27 : "NFS4ERR_FBIG",
       28 : "NFS4ERR_NOSPC",
       30 : "NFS4ERR_ROFS",
       31 : "NFS4ERR_MLINK",
       63 : "NFS4ERR_NAMETOOLONG",
       66 : "NFS4ERR_NOTEMPTY",
       69 : "NFS4ERR_DQUOT",
       70 : "NFS4ERR_STALE",
    10001 : "NFS4ERR_BADHANDLE",
    10003 : "NFS4ERR_BAD_COOKIE",
    10004 : "NFS4ERR_NOTSUPP",
    10005 : "NFS4ERR_TOOSMALL",
    10006 : "NFS4ERR_SERVERFAULT",
    10007 : "NFS4ERR_BADTYPE",
    10008 : "NFS4ERR_DELAY",
    10009 : "NFS4ERR_SAME",
    10010 : "NFS4ERR_DENIED",
    10011 : "NFS4ERR_EXPIRED",
    10012 : "NFS4ERR_LOCKED",
    10013 : "NFS4ERR_GRACE",
    10014 : "NFS4ERR_FHEXPIRED",
    10015 : "NFS4ERR_SHARE_DENIED",
    10016 : "NFS4ERR_WRONGSEC",
    10017 : "NFS4ERR_CLID_INUSE",
    10018 : "NFS4ERR_RESOURCE",
    10019 : "NFS4ERR_MOVED",
    10020 : "NFS4ERR_NOFILEHANDLE",
    10021 : "NFS4ERR_MINOR_VERS_MISMATCH",
    10022 : "NFS4ERR_STALE_CLIENTID",
    10023 : "NFS4ERR_STALE_STATEID",
    10024 : "NFS4ERR_OLD_STATEID",
    10025 : "NFS4ERR_BAD_STATEID",
    10026 : "NFS4ERR_BAD_SEQID",
    10027 : "NFS4ERR_NOT_SAME",
    10028 : "NFS4ERR_LOCK_RANGE",
    10029 : "NFS4ERR_SYMLINK",
    10030 : "NFS4ERR_RESTOREFH",
    10031 : "NFS4ERR_LEASE_MOVED",
    10032 : "NFS4ERR_ATTRNOTSUPP",
    10033 : "NFS4ERR_NO_GRACE",
    10034 : "NFS4ERR_RECLAIM_BAD",
    10035 : "NFS4ERR_RECLAIM_CONFLICT",
    10036 : "NFS4ERR_BADXDR",
    10037 : "NFS4ERR_LOCKS_HELD",
    10038 : "NFS4ERR_OPENMODE",
    10039 : "NFS4ERR_BADOWNER",
    10040 : "NFS4ERR_BADCHAR",
    10041 : "NFS4ERR_BADNAME",
    10042 : "NFS4ERR_BAD_RANGE",
    10043 : "NFS4ERR_LOCK_NOTSUPP",
    10044 : "NFS4ERR_OP_ILLEGAL",
    10045 : "NFS4ERR_DEADLOCK",
    10046 : "NFS4ERR_FILE_OPEN",
    10047 : "NFS4ERR_ADMIN_REVOKED",
    10048 : "NFS4ERR_CB_PATH_DOWN",
    10049 : "NFS4ERR_BADIOMODE",
    10050 : "NFS4ERR_BADLAYOUT",
    10051 : "NFS4ERR_BAD_SESSION_DIGEST",
    10052 : "NFS4ERR_BADSESSION",
    10053 : "NFS4ERR_BADSLOT",
    10054 : "NFS4ERR_COMPLETE_ALREADY",
    10055 : "NFS4ERR_CONN_NOT_BOUND_TO_SESSION",
    10056 : "NFS4ERR_DELEG_ALREADY_WANTED",
    10057 : "NFS4ERR_BACK_CHAN_BUSY",
    10058 : "NFS4ERR_LAYOUTTRYLATER",
    10059 : "NFS4ERR_LAYOUTUNAVAILABLE",
    10060 : "NFS4ERR_NOMATCHING_LAYOUT",
    10061 : "NFS4ERR_RECALLCONFLICT",
    10062 : "NFS4ERR_UNKNOWN_LAYOUTTYPE",
    10063 : "NFS4ERR_SEQ_MISORDERED",
    10064 : "NFS4ERR_SEQUENCE_POS",
    10065 : "NFS4ERR_REQ_TOO_BIG",
    10066 : "NFS4ERR_REP_TOO_BIG",
    10067 : "NFS4ERR_REP_TOO_BIG_TO_CACHE",
    10068 : "NFS4ERR_RETRY_UNCACHED_REP",
    10069 : "NFS4ERR_UNSAFE_COMPOUND",
    10070 : "NFS4ERR_TOO_MANY_OPS",
    10071 : "NFS4ERR_OP_NOT_IN_SESSION",
    10072 : "NFS4ERR_HASH_ALG_UNSUPP",
    10074 : "NFS4ERR_CLIENTID_BUSY",
    10075 : "NFS4ERR_PNFS_IO_HOLE",
    10076 : "NFS4ERR_SEQ_FALSE_RETRY",
    10077 : "NFS4ERR_BAD_HIGH_SLOT",
    10078 : "NFS4ERR_DEADSESSION",
    10079 : "NFS4ERR_ENCR_ALG_UNSUPP",
    10080 : "NFS4ERR_PNFS_NO_LAYOUT",
    10081 : "NFS4ERR_NOT_ONLY_OP",
    10082 : "NFS4ERR_WRONG_CRED",
    10083 : "NFS4ERR_WRONG_TYPE",
    10084 : "NFS4ERR_DIRDELEG_UNAVAIL",
    10085 : "NFS4ERR_REJECT_DELEG",
    10086 : "NFS4ERR_RETURNCONFLICT",
    10087 : "NFS4ERR_DELEG_REVOKED",
    10088 : "NFS4ERR_PARTNER_NOTSUPP",
    10089 : "NFS4ERR_PARTNER_NO_AUTH",
    10090 : "NFS4ERR_UNION_NOTSUPP",
    10091 : "NFS4ERR_OFFLOAD_DENIED",
    10092 : "NFS4ERR_WRONG_LFS",
    10093 : "NFS4ERR_BADLABEL",
    10094 : "NFS4ERR_OFFLOAD_NO_REQS",
}

# Enum time_how4
SET_TO_SERVER_TIME4 = 0
SET_TO_CLIENT_TIME4 = 1

time_how4 = {
    0 : "SET_TO_SERVER_TIME4",
    1 : "SET_TO_CLIENT_TIME4",
}

# Various Access Control Entry definitions
#
# Mask that indicates which Access Control Entries are supported.
# Values for the fattr4_aclsupport attribute.
ACL4_SUPPORT_ALLOW_ACL          = 0x00000001
ACL4_SUPPORT_DENY_ACL           = 0x00000002
ACL4_SUPPORT_AUDIT_ACL          = 0x00000004
ACL4_SUPPORT_ALARM_ACL          = 0x00000008

# acetype4 values, others can be added as needed.
ACE4_ACCESS_ALLOWED_ACE_TYPE    = 0x00000000
ACE4_ACCESS_DENIED_ACE_TYPE     = 0x00000001
ACE4_SYSTEM_AUDIT_ACE_TYPE      = 0x00000002
ACE4_SYSTEM_ALARM_ACE_TYPE      = 0x00000003

# ACE flag values
ACE4_FILE_INHERIT_ACE           = 0x00000001
ACE4_DIRECTORY_INHERIT_ACE      = 0x00000002
ACE4_NO_PROPAGATE_INHERIT_ACE   = 0x00000004
ACE4_INHERIT_ONLY_ACE           = 0x00000008
ACE4_SUCCESSFUL_ACCESS_ACE_FLAG = 0x00000010
ACE4_FAILED_ACCESS_ACE_FLAG     = 0x00000020
ACE4_IDENTIFIER_GROUP           = 0x00000040
ACE4_INHERITED_ACE              = 0x00000080  # New to NFSv4.1

# ACE mask values
ACE4_READ_DATA                  = 0x00000001
ACE4_LIST_DIRECTORY             = 0x00000001
ACE4_WRITE_DATA                 = 0x00000002
ACE4_ADD_FILE                   = 0x00000002
ACE4_APPEND_DATA                = 0x00000004
ACE4_ADD_SUBDIRECTORY           = 0x00000004
ACE4_READ_NAMED_ATTRS           = 0x00000008
ACE4_WRITE_NAMED_ATTRS          = 0x00000010
ACE4_EXECUTE                    = 0x00000020
ACE4_DELETE_CHILD               = 0x00000040
ACE4_READ_ATTRIBUTES            = 0x00000080
ACE4_WRITE_ATTRIBUTES           = 0x00000100
ACE4_WRITE_RETENTION            = 0x00000200  # New to NFSv4.1
ACE4_WRITE_RETENTION_HOLD       = 0x00000400  # New to NFSv4.1
ACE4_DELETE                     = 0x00010000
ACE4_READ_ACL                   = 0x00020000
ACE4_WRITE_ACL                  = 0x00040000
ACE4_WRITE_OWNER                = 0x00080000
ACE4_SYNCHRONIZE                = 0x00100000

# ACE4_GENERIC_READ -- defined as combination of
#      ACE4_READ_ACL |
#      ACE4_READ_DATA |
#      ACE4_READ_ATTRIBUTES |
#      ACE4_SYNCHRONIZE
ACE4_GENERIC_READ               = 0x00120081

# ACE4_GENERIC_WRITE -- defined as combination of
#      ACE4_READ_ACL |
#      ACE4_WRITE_DATA |
#      ACE4_WRITE_ATTRIBUTES |
#      ACE4_WRITE_ACL |
#      ACE4_APPEND_DATA |
#      ACE4_SYNCHRONIZE
ACE4_GENERIC_WRITE              = 0x00160106

# ACE4_GENERIC_EXECUTE -- defined as combination of
#      ACE4_READ_ACL
#      ACE4_READ_ATTRIBUTES
#      ACE4_EXECUTE
#      ACE4_SYNCHRONIZE
ACE4_GENERIC_EXECUTE            = 0x001200A0

# ACL flag values new to NFSv4.1
ACL4_AUTO_INHERIT = 0x00000001
ACL4_PROTECTED    = 0x00000002
ACL4_DEFAULTED    = 0x00000004

# Field definitions for the fattr4_mode attribute
# and fattr4_mode_set_masked attributes.
MODE4_SUID = 0x800  # set user id on execution
MODE4_SGID = 0x400  # set group id on execution
MODE4_SVTX = 0x200  # save text even after use
MODE4_RUSR = 0x100  # read permission: owner
MODE4_WUSR = 0x080  # write permission: owner
MODE4_XUSR = 0x040  # execute permission: owner
MODE4_RGRP = 0x020  # read permission: group
MODE4_WGRP = 0x010  # write permission: group
MODE4_XGRP = 0x008  # execute permission: group
MODE4_ROTH = 0x004  # read permission: other
MODE4_WOTH = 0x002  # write permission: other
MODE4_XOTH = 0x001  # execute permission: other

# Enum stable_how4
UNSTABLE4  = 0
DATA_SYNC4 = 1
FILE_SYNC4 = 2

stable_how4 = {
    0 : "UNSTABLE4",
    1 : "DATA_SYNC4",
    2 : "FILE_SYNC4",
}

# Values for fattr4_fh_expire_type
FH4_PERSISTENT         = 0x00000000
FH4_NOEXPIRE_WITH_OPEN = 0x00000001
FH4_VOLATILE_ANY       = 0x00000002
FH4_VOL_MIGRATION      = 0x00000004
FH4_VOL_RENAME         = 0x00000008

# Enum layouttype4
LAYOUT4_NFSV4_1_FILES = 0x1
LAYOUT4_OSD2_OBJECTS  = 0x2
LAYOUT4_BLOCK_VOLUME  = 0x3
LAYOUT4_FLEX_FILES    = 0x4

layouttype4 = {
    0x1 : "LAYOUT4_NFSV4_1_FILES",
    0x2 : "LAYOUT4_OSD2_OBJECTS",
    0x3 : "LAYOUT4_BLOCK_VOLUME",
    0x4 : "LAYOUT4_FLEX_FILES",
}

NFL4_UFLG_MASK                  = 0x0000003F
NFL4_UFLG_DENSE                 = 0x00000001
NFL4_UFLG_COMMIT_THRU_MDS       = 0x00000002
NFL42_UFLG_IO_ADVISE_THRU_MDS   = 0x00000004
NFL4_UFLG_STRIPE_UNIT_SIZE_MASK = 0xFFFFFFC0

# Enum filelayout_hint_care4
NFLH4_CARE_DENSE              = NFL4_UFLG_DENSE
NFLH4_CARE_COMMIT_THRU_MDS    = NFL4_UFLG_COMMIT_THRU_MDS
NFL42_CARE_IO_ADVISE_THRU_MDS = NFL42_UFLG_IO_ADVISE_THRU_MDS
NFLH4_CARE_STRIPE_UNIT_SIZE   = 0x00000040
NFLH4_CARE_STRIPE_COUNT       = 0x00000080

filelayout_hint_care4 = {
                  NFL4_UFLG_DENSE : "NFLH4_CARE_DENSE",
        NFL4_UFLG_COMMIT_THRU_MDS : "NFLH4_CARE_COMMIT_THRU_MDS",
    NFL42_UFLG_IO_ADVISE_THRU_MDS : "NFL42_CARE_IO_ADVISE_THRU_MDS",
                       0x00000040 : "NFLH4_CARE_STRIPE_UNIT_SIZE",
                       0x00000080 : "NFLH4_CARE_STRIPE_COUNT",
}

# NFSv4.x flex files layout definitions (BEGIN) ================================

FF_FLAGS_NO_LAYOUTCOMMIT = 0x00000001
FF_FLAGS_NO_IO_THRU_MDS  = 0x00000002
FF_FLAGS_NO_READ_IO      = 0x00000004

# Enum ff_cb_recall_any_mask
FF_RCA4_TYPE_MASK_READ = -2
FF_RCA4_TYPE_MASK_RW   = -1

ff_cb_recall_any_mask = {
    -2 : "FF_RCA4_TYPE_MASK_READ",
    -1 : "FF_RCA4_TYPE_MASK_RW",
}

# NFSv4.x flex files layout definitions (END) ==================================

# Enum layoutiomode4
LAYOUTIOMODE4_READ = 1
LAYOUTIOMODE4_RW   = 2
LAYOUTIOMODE4_ANY  = 3

layoutiomode4 = {
    1 : "LAYOUTIOMODE4_READ",
    2 : "LAYOUTIOMODE4_RW",
    3 : "LAYOUTIOMODE4_ANY",
}
# Constants used for LAYOUTRETURN and CB_LAYOUTRECALL
LAYOUT4_RET_REC_FILE = 1
LAYOUT4_RET_REC_FSID = 2
LAYOUT4_RET_REC_ALL  = 3

# Enum layoutreturn_type4
LAYOUTRETURN4_FILE = LAYOUT4_RET_REC_FILE
LAYOUTRETURN4_FSID = LAYOUT4_RET_REC_FSID
LAYOUTRETURN4_ALL  = LAYOUT4_RET_REC_ALL

layoutreturn_type4 = {
    LAYOUT4_RET_REC_FILE : "LAYOUTRETURN4_FILE",
    LAYOUT4_RET_REC_FSID : "LAYOUTRETURN4_FSID",
     LAYOUT4_RET_REC_ALL : "LAYOUTRETURN4_ALL",
}

# Enum fs4_status_type
STATUS4_FIXED     = 1
STATUS4_UPDATED   = 2
STATUS4_VERSIONED = 3
STATUS4_WRITABLE  = 4
STATUS4_REFERRAL  = 5

fs4_status_type = {
    1 : "STATUS4_FIXED",
    2 : "STATUS4_UPDATED",
    3 : "STATUS4_VERSIONED",
    4 : "STATUS4_WRITABLE",
    5 : "STATUS4_REFERRAL",
}

TH4_READ_SIZE    = 0
TH4_WRITE_SIZE   = 1
TH4_READ_IOSIZE  = 2
TH4_WRITE_IOSIZE = 3

RET4_DURATION_INFINITE = 0xffffffffffffffff

# Byte indices of items within
# fls_info: flag fields, class numbers,
# bytes indicating ranks and orders.
FSLI4BX_GFLAGS     = 0
FSLI4BX_TFLAGS     = 1
FSLI4BX_CLSIMUL    = 2
FSLI4BX_CLHANDLE   = 3
FSLI4BX_CLFILEID   = 4
FSLI4BX_CLWRITEVER = 5
FSLI4BX_CLCHANGE   = 6
FSLI4BX_CLREADDIR  = 7
FSLI4BX_READRANK   = 8
FSLI4BX_WRITERANK  = 9
FSLI4BX_READORDER  = 10
FSLI4BX_WRITEORDER = 11

# Bits defined within the general flag byte.
FSLI4GF_WRITABLE   = 0x01
FSLI4GF_CUR_REQ    = 0x02
FSLI4GF_ABSENT     = 0x04
FSLI4GF_GOING      = 0x08
FSLI4GF_SPLIT      = 0x10

# Bits defined within the transport flag byte.
FSLI4TF_RDMA       = 0x01

# Flag bits in fli_flags.
FSLI4IF_VAR_SUB    = 0x00000001
# Constants for fs_charset_cap4
FSCHARSET_CAP4_CONTAINS_NON_UTF8 = 0x1
FSCHARSET_CAP4_ALLOWS_ONLY_UTF8  = 0x2

# Enum netloc_type4
NL4_NAME    = 1
NL4_URL     = 2
NL4_NETADDR = 3

netloc_type4 = {
    1 : "NL4_NAME",
    2 : "NL4_URL",
    3 : "NL4_NETADDR",
}

# Enum change_attr_type4
NFS4_CHANGE_TYPE_IS_MONOTONIC_INCR         = 0
NFS4_CHANGE_TYPE_IS_VERSION_COUNTER        = 1
NFS4_CHANGE_TYPE_IS_VERSION_COUNTER_NOPNFS = 2
NFS4_CHANGE_TYPE_IS_TIME_METADATA          = 3
NFS4_CHANGE_TYPE_IS_UNDEFINED              = 4

change_attr_type4 = {
    0 : "NFS4_CHANGE_TYPE_IS_MONOTONIC_INCR",
    1 : "NFS4_CHANGE_TYPE_IS_VERSION_COUNTER",
    2 : "NFS4_CHANGE_TYPE_IS_VERSION_COUNTER_NOPNFS",
    3 : "NFS4_CHANGE_TYPE_IS_TIME_METADATA",
    4 : "NFS4_CHANGE_TYPE_IS_UNDEFINED",
}

# Enum nfs_fattr4

# Mandatory Attributes
FATTR4_SUPPORTED_ATTRS    = 0
FATTR4_TYPE               = 1
FATTR4_FH_EXPIRE_TYPE     = 2
FATTR4_CHANGE             = 3
FATTR4_SIZE               = 4
FATTR4_LINK_SUPPORT       = 5
FATTR4_SYMLINK_SUPPORT    = 6
FATTR4_NAMED_ATTR         = 7
FATTR4_FSID               = 8
FATTR4_UNIQUE_HANDLES     = 9
FATTR4_LEASE_TIME         = 10
FATTR4_RDATTR_ERROR       = 11
FATTR4_FILEHANDLE         = 19
FATTR4_SUPPATTR_EXCLCREAT = 75  # New to NFSv4.1

# Recommended Attributes
FATTR4_ACL                = 12
FATTR4_ACLSUPPORT         = 13
FATTR4_ARCHIVE            = 14
FATTR4_CANSETTIME         = 15
FATTR4_CASE_INSENSITIVE   = 16
FATTR4_CASE_PRESERVING    = 17
FATTR4_CHOWN_RESTRICTED   = 18
FATTR4_FILEID             = 20
FATTR4_FILES_AVAIL        = 21
FATTR4_FILES_FREE         = 22
FATTR4_FILES_TOTAL        = 23
FATTR4_FS_LOCATIONS       = 24
FATTR4_HIDDEN             = 25
FATTR4_HOMOGENEOUS        = 26
FATTR4_MAXFILESIZE        = 27
FATTR4_MAXLINK            = 28
FATTR4_MAXNAME            = 29
FATTR4_MAXREAD            = 30
FATTR4_MAXWRITE           = 31
FATTR4_MIMETYPE           = 32
FATTR4_MODE               = 33
FATTR4_NO_TRUNC           = 34
FATTR4_NUMLINKS           = 35
FATTR4_OWNER              = 36
FATTR4_OWNER_GROUP        = 37
FATTR4_QUOTA_AVAIL_HARD   = 38
FATTR4_QUOTA_AVAIL_SOFT   = 39
FATTR4_QUOTA_USED         = 40
FATTR4_RAWDEV             = 41
FATTR4_SPACE_AVAIL        = 42
FATTR4_SPACE_FREE         = 43
FATTR4_SPACE_TOTAL        = 44
FATTR4_SPACE_USED         = 45
FATTR4_SYSTEM             = 46
FATTR4_TIME_ACCESS        = 47
FATTR4_TIME_ACCESS_SET    = 48
FATTR4_TIME_BACKUP        = 49
FATTR4_TIME_CREATE        = 50
FATTR4_TIME_DELTA         = 51
FATTR4_TIME_METADATA      = 52
FATTR4_TIME_MODIFY        = 53
FATTR4_TIME_MODIFY_SET    = 54
FATTR4_MOUNTED_ON_FILEID  = 55

# New to NFSv4.1
FATTR4_DIR_NOTIF_DELAY    = 56
FATTR4_DIRENT_NOTIF_DELAY = 57
FATTR4_DACL               = 58
FATTR4_SACL               = 59
FATTR4_CHANGE_POLICY      = 60
FATTR4_FS_STATUS          = 61
FATTR4_FS_LAYOUT_TYPES    = 62
FATTR4_LAYOUT_HINT        = 63
FATTR4_LAYOUT_TYPES       = 64
FATTR4_LAYOUT_BLKSIZE     = 65
FATTR4_LAYOUT_ALIGNMENT   = 66
FATTR4_FS_LOCATIONS_INFO  = 67
FATTR4_MDSTHRESHOLD       = 68
FATTR4_RETENTION_GET      = 69
FATTR4_RETENTION_SET      = 70
FATTR4_RETENTEVT_GET      = 71
FATTR4_RETENTEVT_SET      = 72
FATTR4_RETENTION_HOLD     = 73
FATTR4_MODE_SET_MASKED    = 74
FATTR4_FS_CHARSET_CAP     = 76

# New to NFSv4.2
FATTR4_CLONE_BLKSIZE      = 77
FATTR4_SPACE_FREED        = 78
FATTR4_CHANGE_ATTR_TYPE   = 79
FATTR4_SEC_LABEL          = 80
FATTR4_MODE_UMASK         = 81  # draft-bfields-nfsv4-umask-01

nfs_fattr4 = {
     0 : "FATTR4_SUPPORTED_ATTRS",
     1 : "FATTR4_TYPE",
     2 : "FATTR4_FH_EXPIRE_TYPE",
     3 : "FATTR4_CHANGE",
     4 : "FATTR4_SIZE",
     5 : "FATTR4_LINK_SUPPORT",
     6 : "FATTR4_SYMLINK_SUPPORT",
     7 : "FATTR4_NAMED_ATTR",
     8 : "FATTR4_FSID",
     9 : "FATTR4_UNIQUE_HANDLES",
    10 : "FATTR4_LEASE_TIME",
    11 : "FATTR4_RDATTR_ERROR",
    19 : "FATTR4_FILEHANDLE",
    75 : "FATTR4_SUPPATTR_EXCLCREAT",
    12 : "FATTR4_ACL",
    13 : "FATTR4_ACLSUPPORT",
    14 : "FATTR4_ARCHIVE",
    15 : "FATTR4_CANSETTIME",
    16 : "FATTR4_CASE_INSENSITIVE",
    17 : "FATTR4_CASE_PRESERVING",
    18 : "FATTR4_CHOWN_RESTRICTED",
    20 : "FATTR4_FILEID",
    21 : "FATTR4_FILES_AVAIL",
    22 : "FATTR4_FILES_FREE",
    23 : "FATTR4_FILES_TOTAL",
    24 : "FATTR4_FS_LOCATIONS",
    25 : "FATTR4_HIDDEN",
    26 : "FATTR4_HOMOGENEOUS",
    27 : "FATTR4_MAXFILESIZE",
    28 : "FATTR4_MAXLINK",
    29 : "FATTR4_MAXNAME",
    30 : "FATTR4_MAXREAD",
    31 : "FATTR4_MAXWRITE",
    32 : "FATTR4_MIMETYPE",
    33 : "FATTR4_MODE",
    34 : "FATTR4_NO_TRUNC",
    35 : "FATTR4_NUMLINKS",
    36 : "FATTR4_OWNER",
    37 : "FATTR4_OWNER_GROUP",
    38 : "FATTR4_QUOTA_AVAIL_HARD",
    39 : "FATTR4_QUOTA_AVAIL_SOFT",
    40 : "FATTR4_QUOTA_USED",
    41 : "FATTR4_RAWDEV",
    42 : "FATTR4_SPACE_AVAIL",
    43 : "FATTR4_SPACE_FREE",
    44 : "FATTR4_SPACE_TOTAL",
    45 : "FATTR4_SPACE_USED",
    46 : "FATTR4_SYSTEM",
    47 : "FATTR4_TIME_ACCESS",
    48 : "FATTR4_TIME_ACCESS_SET",
    49 : "FATTR4_TIME_BACKUP",
    50 : "FATTR4_TIME_CREATE",
    51 : "FATTR4_TIME_DELTA",
    52 : "FATTR4_TIME_METADATA",
    53 : "FATTR4_TIME_MODIFY",
    54 : "FATTR4_TIME_MODIFY_SET",
    55 : "FATTR4_MOUNTED_ON_FILEID",
    56 : "FATTR4_DIR_NOTIF_DELAY",
    57 : "FATTR4_DIRENT_NOTIF_DELAY",
    58 : "FATTR4_DACL",
    59 : "FATTR4_SACL",
    60 : "FATTR4_CHANGE_POLICY",
    61 : "FATTR4_FS_STATUS",
    62 : "FATTR4_FS_LAYOUT_TYPES",
    63 : "FATTR4_LAYOUT_HINT",
    64 : "FATTR4_LAYOUT_TYPES",
    65 : "FATTR4_LAYOUT_BLKSIZE",
    66 : "FATTR4_LAYOUT_ALIGNMENT",
    67 : "FATTR4_FS_LOCATIONS_INFO",
    68 : "FATTR4_MDSTHRESHOLD",
    69 : "FATTR4_RETENTION_GET",
    70 : "FATTR4_RETENTION_SET",
    71 : "FATTR4_RETENTEVT_GET",
    72 : "FATTR4_RETENTEVT_SET",
    73 : "FATTR4_RETENTION_HOLD",
    74 : "FATTR4_MODE_SET_MASKED",
    76 : "FATTR4_FS_CHARSET_CAP",
    77 : "FATTR4_CLONE_BLKSIZE",
    78 : "FATTR4_SPACE_FREED",
    79 : "FATTR4_CHANGE_ATTR_TYPE",
    80 : "FATTR4_SEC_LABEL",
    81 : "FATTR4_MODE_UMASK",
}

# Enum ssv_subkey4
SSV4_SUBKEY_MIC_I2T  = 1
SSV4_SUBKEY_MIC_T2I  = 2
SSV4_SUBKEY_SEAL_I2T = 3
SSV4_SUBKEY_SEAL_T2I = 4

ssv_subkey4 = {
    1 : "SSV4_SUBKEY_MIC_I2T",
    2 : "SSV4_SUBKEY_MIC_T2I",
    3 : "SSV4_SUBKEY_SEAL_I2T",
    4 : "SSV4_SUBKEY_SEAL_T2I",
}

# Enum nfs_opnum4
OP_ACCESS               = 3
OP_CLOSE                = 4
OP_COMMIT               = 5
OP_CREATE               = 6
OP_DELEGPURGE           = 7
OP_DELEGRETURN          = 8
OP_GETATTR              = 9
OP_GETFH                = 10
OP_LINK                 = 11
OP_LOCK                 = 12
OP_LOCKT                = 13
OP_LOCKU                = 14
OP_LOOKUP               = 15
OP_LOOKUPP              = 16
OP_NVERIFY              = 17
OP_OPEN                 = 18
OP_OPENATTR             = 19
OP_OPEN_CONFIRM         = 20     # Mandatory not-to-implement in NFSv4.1
OP_OPEN_DOWNGRADE       = 21
OP_PUTFH                = 22
OP_PUTPUBFH             = 23
OP_PUTROOTFH            = 24
OP_READ                 = 25
OP_READDIR              = 26
OP_READLINK             = 27
OP_REMOVE               = 28
OP_RENAME               = 29
OP_RENEW                = 30     # Mandatory not-to-implement in NFSv4.1
OP_RESTOREFH            = 31
OP_SAVEFH               = 32
OP_SECINFO              = 33
OP_SETATTR              = 34
OP_SETCLIENTID          = 35     # Mandatory not-to-implement in NFSv4.1
OP_SETCLIENTID_CONFIRM  = 36     # Mandatory not-to-implement in NFSv4.1
OP_VERIFY               = 37
OP_WRITE                = 38
OP_RELEASE_LOCKOWNER    = 39     # Mandatory not-to-implement in NFSv4.1
# New operations for NFSv4.1
OP_BACKCHANNEL_CTL      = 40
OP_BIND_CONN_TO_SESSION = 41
OP_EXCHANGE_ID          = 42
OP_CREATE_SESSION       = 43
OP_DESTROY_SESSION      = 44
OP_FREE_STATEID         = 45
OP_GET_DIR_DELEGATION   = 46
OP_GETDEVICEINFO        = 47
OP_GETDEVICELIST        = 48     # Mandatory not-to-implement in NFSv4.2
OP_LAYOUTCOMMIT         = 49
OP_LAYOUTGET            = 50
OP_LAYOUTRETURN         = 51
OP_SECINFO_NO_NAME      = 52
OP_SEQUENCE             = 53
OP_SET_SSV              = 54
OP_TEST_STATEID         = 55
OP_WANT_DELEGATION      = 56
OP_DESTROY_CLIENTID     = 57
OP_RECLAIM_COMPLETE     = 58
# New operations for NFSv4.2
OP_ALLOCATE             = 59
OP_COPY                 = 60
OP_COPY_NOTIFY          = 61
OP_DEALLOCATE           = 62
OP_IO_ADVISE            = 63
OP_LAYOUTERROR          = 64
OP_LAYOUTSTATS          = 65
OP_OFFLOAD_CANCEL       = 66
OP_OFFLOAD_STATUS       = 67
OP_READ_PLUS            = 68
OP_SEEK                 = 69
OP_WRITE_SAME           = 70
OP_CLONE                = 71
# Illegal operation
OP_ILLEGAL              = 10044

nfs_opnum4 = {
        3 : "OP_ACCESS",
        4 : "OP_CLOSE",
        5 : "OP_COMMIT",
        6 : "OP_CREATE",
        7 : "OP_DELEGPURGE",
        8 : "OP_DELEGRETURN",
        9 : "OP_GETATTR",
       10 : "OP_GETFH",
       11 : "OP_LINK",
       12 : "OP_LOCK",
       13 : "OP_LOCKT",
       14 : "OP_LOCKU",
       15 : "OP_LOOKUP",
       16 : "OP_LOOKUPP",
       17 : "OP_NVERIFY",
       18 : "OP_OPEN",
       19 : "OP_OPENATTR",
       20 : "OP_OPEN_CONFIRM",
       21 : "OP_OPEN_DOWNGRADE",
       22 : "OP_PUTFH",
       23 : "OP_PUTPUBFH",
       24 : "OP_PUTROOTFH",
       25 : "OP_READ",
       26 : "OP_READDIR",
       27 : "OP_READLINK",
       28 : "OP_REMOVE",
       29 : "OP_RENAME",
       30 : "OP_RENEW",
       31 : "OP_RESTOREFH",
       32 : "OP_SAVEFH",
       33 : "OP_SECINFO",
       34 : "OP_SETATTR",
       35 : "OP_SETCLIENTID",
       36 : "OP_SETCLIENTID_CONFIRM",
       37 : "OP_VERIFY",
       38 : "OP_WRITE",
       39 : "OP_RELEASE_LOCKOWNER",
       40 : "OP_BACKCHANNEL_CTL",
       41 : "OP_BIND_CONN_TO_SESSION",
       42 : "OP_EXCHANGE_ID",
       43 : "OP_CREATE_SESSION",
       44 : "OP_DESTROY_SESSION",
       45 : "OP_FREE_STATEID",
       46 : "OP_GET_DIR_DELEGATION",
       47 : "OP_GETDEVICEINFO",
       48 : "OP_GETDEVICELIST",
       49 : "OP_LAYOUTCOMMIT",
       50 : "OP_LAYOUTGET",
       51 : "OP_LAYOUTRETURN",
       52 : "OP_SECINFO_NO_NAME",
       53 : "OP_SEQUENCE",
       54 : "OP_SET_SSV",
       55 : "OP_TEST_STATEID",
       56 : "OP_WANT_DELEGATION",
       57 : "OP_DESTROY_CLIENTID",
       58 : "OP_RECLAIM_COMPLETE",
       59 : "OP_ALLOCATE",
       60 : "OP_COPY",
       61 : "OP_COPY_NOTIFY",
       62 : "OP_DEALLOCATE",
       63 : "OP_IO_ADVISE",
       64 : "OP_LAYOUTERROR",
       65 : "OP_LAYOUTSTATS",
       66 : "OP_OFFLOAD_CANCEL",
       67 : "OP_OFFLOAD_STATUS",
       68 : "OP_READ_PLUS",
       69 : "OP_SEEK",
       70 : "OP_WRITE_SAME",
       71 : "OP_CLONE",
    10044 : "OP_ILLEGAL",
}

ACCESS4_READ    = 0x00000001
ACCESS4_LOOKUP  = 0x00000002
ACCESS4_MODIFY  = 0x00000004
ACCESS4_EXTEND  = 0x00000008
ACCESS4_DELETE  = 0x00000010
ACCESS4_EXECUTE = 0x00000020

# Enum nfs_lock_type4
READ_LT   = 1
WRITE_LT  = 2
READW_LT  = 3  # blocking read
WRITEW_LT = 4  # blocking write

nfs_lock_type4 = {
    1 : "READ_LT",
    2 : "WRITE_LT",
    3 : "READW_LT",
    4 : "WRITEW_LT",
}

# Enum createmode4
UNCHECKED4   = 0
GUARDED4     = 1
# Deprecated in NFSv4.1.
EXCLUSIVE4   = 2

# New to NFSv4.1. If session is persistent,
# GUARDED4 MUST be used. Otherwise, use
# EXCLUSIVE4_1 instead of EXCLUSIVE4.
EXCLUSIVE4_1 = 3

createmode4 = {
    0 : "UNCHECKED4",
    1 : "GUARDED4",
    2 : "EXCLUSIVE4",
    3 : "EXCLUSIVE4_1",
}

# Enum opentype4
OPEN4_NOCREATE = 0
OPEN4_CREATE   = 1

opentype4 = {
    0 : "OPEN4_NOCREATE",
    1 : "OPEN4_CREATE",
}

# Enum limit_by4
NFS_LIMIT_SIZE   = 1
NFS_LIMIT_BLOCKS = 2

limit_by4 = {
    1 : "NFS_LIMIT_SIZE",
    2 : "NFS_LIMIT_BLOCKS",
}

# Share Access and Deny constants for open argument
OPEN4_SHARE_ACCESS_READ                               = 0x00000001
OPEN4_SHARE_ACCESS_WRITE                              = 0x00000002
OPEN4_SHARE_ACCESS_BOTH                               = 0x00000003
OPEN4_SHARE_DENY_NONE                                 = 0x00000000
OPEN4_SHARE_DENY_READ                                 = 0x00000001
OPEN4_SHARE_DENY_WRITE                                = 0x00000002
OPEN4_SHARE_DENY_BOTH                                 = 0x00000003
# New flags for share_access field of OPEN4args
OPEN4_SHARE_ACCESS_WANT_DELEG_MASK                    = 0xFF00
OPEN4_SHARE_ACCESS_WANT_NO_PREFERENCE                 = 0x0000
OPEN4_SHARE_ACCESS_WANT_READ_DELEG                    = 0x0100
OPEN4_SHARE_ACCESS_WANT_WRITE_DELEG                   = 0x0200
OPEN4_SHARE_ACCESS_WANT_ANY_DELEG                     = 0x0300
OPEN4_SHARE_ACCESS_WANT_NO_DELEG                      = 0x0400
OPEN4_SHARE_ACCESS_WANT_CANCEL                        = 0x0500
OPEN4_SHARE_ACCESS_WANT_SIGNAL_DELEG_WHEN_RESRC_AVAIL = 0x10000
OPEN4_SHARE_ACCESS_WANT_PUSH_DELEG_WHEN_UNCONTENDED   = 0x20000

# Enum open_delegation_type4
OPEN_DELEGATE_NONE     = 0
OPEN_DELEGATE_READ     = 1
OPEN_DELEGATE_WRITE    = 2
OPEN_DELEGATE_NONE_EXT = 3  # New to NFSv4.1

open_delegation_type4 = {
    0 : "OPEN_DELEGATE_NONE",
    1 : "OPEN_DELEGATE_READ",
    2 : "OPEN_DELEGATE_WRITE",
    3 : "OPEN_DELEGATE_NONE_EXT",
}

# Enum open_claim_type4

# Not a reclaim.
CLAIM_NULL          = 0
CLAIM_PREVIOUS      = 1
CLAIM_DELEGATE_CUR  = 2
CLAIM_DELEGATE_PREV = 3

# Not a reclaim.
# Like CLAIM_NULL, but object identified
# by the current filehandle.
CLAIM_FH            = 4  # New to NFSv4.1

# Like CLAIM_DELEGATE_CUR, but object identified
# by current filehandle.
CLAIM_DELEG_CUR_FH  = 5  # New to NFSv4.1

# Like CLAIM_DELEGATE_PREV, but object identified
# by current filehandle.
CLAIM_DELEG_PREV_FH = 6  # New to NFSv4.1

open_claim_type4 = {
    0 : "CLAIM_NULL",
    1 : "CLAIM_PREVIOUS",
    2 : "CLAIM_DELEGATE_CUR",
    3 : "CLAIM_DELEGATE_PREV",
    4 : "CLAIM_FH",
    5 : "CLAIM_DELEG_CUR_FH",
    6 : "CLAIM_DELEG_PREV_FH",
}

# Enum why_no_delegation4
WND4_NOT_WANTED                 = 0
WND4_CONTENTION                 = 1
WND4_RESOURCE                   = 2
WND4_NOT_SUPP_FTYPE             = 3
WND4_WRITE_DELEG_NOT_SUPP_FTYPE = 4
WND4_NOT_SUPP_UPGRADE           = 5
WND4_NOT_SUPP_DOWNGRADE         = 6
WND4_CANCELLED                  = 7
WND4_IS_DIR                     = 8

why_no_delegation4 = {
    0 : "WND4_NOT_WANTED",
    1 : "WND4_CONTENTION",
    2 : "WND4_RESOURCE",
    3 : "WND4_NOT_SUPP_FTYPE",
    4 : "WND4_WRITE_DELEG_NOT_SUPP_FTYPE",
    5 : "WND4_NOT_SUPP_UPGRADE",
    6 : "WND4_NOT_SUPP_DOWNGRADE",
    7 : "WND4_CANCELLED",
    8 : "WND4_IS_DIR",
}
# Result flags
#
# Client must confirm open
OPEN4_RESULT_CONFIRM           = 0x00000002
# Type of file locking behavior at the server
OPEN4_RESULT_LOCKTYPE_POSIX    = 0x00000004
# Server will preserve file if removed while open
OPEN4_RESULT_PRESERVE_UNLINKED = 0x00000008
# Server may use CB_NOTIFY_LOCK on locks derived from this open
OPEN4_RESULT_MAY_NOTIFY_LOCK   = 0x00000020

# Enum nfs_secflavor4
AUTH_NONE  = 0
AUTH_SYS   = 1
RPCSEC_GSS = 6

nfs_secflavor4 = {
    0 : "AUTH_NONE",
    1 : "AUTH_SYS",
    6 : "RPCSEC_GSS",
}

# Enum rpc_gss_svc_t
RPC_GSS_SVC_NONE      = 1
RPC_GSS_SVC_INTEGRITY = 2
RPC_GSS_SVC_PRIVACY   = 3

rpc_gss_svc_t = {
    1 : "RPC_GSS_SVC_NONE",
    2 : "RPC_GSS_SVC_INTEGRITY",
    3 : "RPC_GSS_SVC_PRIVACY",
}

# Enum channel_dir_from_client4
CDFC4_FORE         = 0x1
CDFC4_BACK         = 0x2
CDFC4_FORE_OR_BOTH = 0x3
CDFC4_BACK_OR_BOTH = 0x7

channel_dir_from_client4 = {
    0x1 : "CDFC4_FORE",
    0x2 : "CDFC4_BACK",
    0x3 : "CDFC4_FORE_OR_BOTH",
    0x7 : "CDFC4_BACK_OR_BOTH",
}

# Enum channel_dir_from_server4
CDFS4_FORE = 0x1
CDFS4_BACK = 0x2
CDFS4_BOTH = 0x3

channel_dir_from_server4 = {
    0x1 : "CDFS4_FORE",
    0x2 : "CDFS4_BACK",
    0x3 : "CDFS4_BOTH",
}

# EXCHANGE_ID: Instantiate Client ID
# ======================================================================
EXCHGID4_FLAG_SUPP_MOVED_REFER    = 0x00000001
EXCHGID4_FLAG_SUPP_MOVED_MIGR     = 0x00000002
EXCHGID4_FLAG_SUPP_FENCE_OPS      = 0x00000004  # New to NFSv4.2
EXCHGID4_FLAG_BIND_PRINC_STATEID  = 0x00000100
EXCHGID4_FLAG_USE_NON_PNFS        = 0x00010000
EXCHGID4_FLAG_USE_PNFS_MDS        = 0x00020000
EXCHGID4_FLAG_USE_PNFS_DS         = 0x00040000
EXCHGID4_FLAG_MASK_PNFS           = 0x00070000
EXCHGID4_FLAG_UPD_CONFIRMED_REC_A = 0x40000000
EXCHGID4_FLAG_CONFIRMED_R         = 0x80000000

# Enum state_protect_how4
SP4_NONE      = 0
SP4_MACH_CRED = 1
SP4_SSV       = 2

state_protect_how4 = {
    0 : "SP4_NONE",
    1 : "SP4_MACH_CRED",
    2 : "SP4_SSV",
}

CREATE_SESSION4_FLAG_PERSIST        = 0x00000001
CREATE_SESSION4_FLAG_CONN_BACK_CHAN = 0x00000002
CREATE_SESSION4_FLAG_CONN_RDMA      = 0x00000004

# Enum gddrnf4_status
GDD4_OK      = 0
GDD4_UNAVAIL = 1

gddrnf4_status = {
    0 : "GDD4_OK",
    1 : "GDD4_UNAVAIL",
}

# Enum secinfo_style4
SECINFO_STYLE4_CURRENT_FH = 0
SECINFO_STYLE4_PARENT     = 1

secinfo_style4 = {
    0 : "SECINFO_STYLE4_CURRENT_FH",
    1 : "SECINFO_STYLE4_PARENT",
}

SEQ4_STATUS_CB_PATH_DOWN               = 0x00000001
SEQ4_STATUS_CB_GSS_CONTEXTS_EXPIRING   = 0x00000002
SEQ4_STATUS_CB_GSS_CONTEXTS_EXPIRED    = 0x00000004
SEQ4_STATUS_EXPIRED_ALL_STATE_REVOKED  = 0x00000008
SEQ4_STATUS_EXPIRED_SOME_STATE_REVOKED = 0x00000010
SEQ4_STATUS_ADMIN_STATE_REVOKED        = 0x00000020
SEQ4_STATUS_RECALLABLE_STATE_REVOKED   = 0x00000040
SEQ4_STATUS_LEASE_MOVED                = 0x00000080
SEQ4_STATUS_RESTART_RECLAIM_NEEDED     = 0x00000100
SEQ4_STATUS_CB_PATH_DOWN_SESSION       = 0x00000200
SEQ4_STATUS_BACKCHANNEL_FAULT          = 0x00000400
SEQ4_STATUS_DEVID_CHANGED              = 0x00000800
SEQ4_STATUS_DEVID_DELETED              = 0x00001000

# Enum IO_ADVISE_type4
IO_ADVISE4_NORMAL                 = 0
IO_ADVISE4_SEQUENTIAL             = 1
IO_ADVISE4_SEQUENTIAL_BACKWARDS   = 2
IO_ADVISE4_RANDOM                 = 3
IO_ADVISE4_WILLNEED               = 4
IO_ADVISE4_WILLNEED_OPPORTUNISTIC = 5
IO_ADVISE4_DONTNEED               = 6
IO_ADVISE4_NOREUSE                = 7
IO_ADVISE4_READ                   = 8
IO_ADVISE4_WRITE                  = 9
IO_ADVISE4_INIT_PROXIMITY         = 10

IO_ADVISE_type4 = {
     0 : "IO_ADVISE4_NORMAL",
     1 : "IO_ADVISE4_SEQUENTIAL",
     2 : "IO_ADVISE4_SEQUENTIAL_BACKWARDS",
     3 : "IO_ADVISE4_RANDOM",
     4 : "IO_ADVISE4_WILLNEED",
     5 : "IO_ADVISE4_WILLNEED_OPPORTUNISTIC",
     6 : "IO_ADVISE4_DONTNEED",
     7 : "IO_ADVISE4_NOREUSE",
     8 : "IO_ADVISE4_READ",
     9 : "IO_ADVISE4_WRITE",
    10 : "IO_ADVISE4_INIT_PROXIMITY",
}

# Enum data_content4
NFS4_CONTENT_DATA = 0
NFS4_CONTENT_HOLE = 1

data_content4 = {
    0 : "NFS4_CONTENT_DATA",
    1 : "NFS4_CONTENT_HOLE",
}

# Enum nfs_cb_opnum4
OP_CB_GETATTR              = 3
OP_CB_RECALL               = 4
# Callback operations new to NFSv4.1
OP_CB_LAYOUTRECALL         = 5
OP_CB_NOTIFY               = 6
OP_CB_PUSH_DELEG           = 7
OP_CB_RECALL_ANY           = 8
OP_CB_RECALLABLE_OBJ_AVAIL = 9
OP_CB_RECALL_SLOT          = 10
OP_CB_SEQUENCE             = 11
OP_CB_WANTS_CANCELLED      = 12
OP_CB_NOTIFY_LOCK          = 13
OP_CB_NOTIFY_DEVICEID      = 14
# Callback operations new to NFSv4.2
OP_CB_OFFLOAD              = 15
# Illegal callback operation
OP_CB_ILLEGAL              = 10044

nfs_cb_opnum4 = {
        3 : "OP_CB_GETATTR",
        4 : "OP_CB_RECALL",
        5 : "OP_CB_LAYOUTRECALL",
        6 : "OP_CB_NOTIFY",
        7 : "OP_CB_PUSH_DELEG",
        8 : "OP_CB_RECALL_ANY",
        9 : "OP_CB_RECALLABLE_OBJ_AVAIL",
       10 : "OP_CB_RECALL_SLOT",
       11 : "OP_CB_SEQUENCE",
       12 : "OP_CB_WANTS_CANCELLED",
       13 : "OP_CB_NOTIFY_LOCK",
       14 : "OP_CB_NOTIFY_DEVICEID",
       15 : "OP_CB_OFFLOAD",
    10044 : "OP_CB_ILLEGAL",
}

# Enum layoutrecall_type4
LAYOUTRECALL4_FILE = LAYOUT4_RET_REC_FILE
LAYOUTRECALL4_FSID = LAYOUT4_RET_REC_FSID
LAYOUTRECALL4_ALL  = LAYOUT4_RET_REC_ALL

layoutrecall_type4 = {
    LAYOUT4_RET_REC_FILE : "LAYOUTRECALL4_FILE",
    LAYOUT4_RET_REC_FSID : "LAYOUTRECALL4_FSID",
     LAYOUT4_RET_REC_ALL : "LAYOUTRECALL4_ALL",
}

# Enum notify_type4
NOTIFY4_CHANGE_CHILD_ATTRS     = 0
NOTIFY4_CHANGE_DIR_ATTRS       = 1
NOTIFY4_REMOVE_ENTRY           = 2
NOTIFY4_ADD_ENTRY              = 3
NOTIFY4_RENAME_ENTRY           = 4
NOTIFY4_CHANGE_COOKIE_VERIFIER = 5

notify_type4 = {
    0 : "NOTIFY4_CHANGE_CHILD_ATTRS",
    1 : "NOTIFY4_CHANGE_DIR_ATTRS",
    2 : "NOTIFY4_REMOVE_ENTRY",
    3 : "NOTIFY4_ADD_ENTRY",
    4 : "NOTIFY4_RENAME_ENTRY",
    5 : "NOTIFY4_CHANGE_COOKIE_VERIFIER",
}

# CB_RECALL_ANY: Keep Any N Recallable Objects
# ======================================================================
RCA4_TYPE_MASK_RDATA_DLG        = 0
RCA4_TYPE_MASK_WDATA_DLG        = 1
RCA4_TYPE_MASK_DIR_DLG          = 2
RCA4_TYPE_MASK_FILE_LAYOUT      = 3
RCA4_TYPE_MASK_BLK_LAYOUT       = 4
RCA4_TYPE_MASK_OBJ_LAYOUT_MIN   = 8
RCA4_TYPE_MASK_OBJ_LAYOUT_MAX   = 9
RCA4_TYPE_MASK_OTHER_LAYOUT_MIN = 12
RCA4_TYPE_MASK_OTHER_LAYOUT_MAX = 15

# Enum notify_deviceid_type4
NOTIFY_DEVICEID4_CHANGE = 1
NOTIFY_DEVICEID4_DELETE = 2

notify_deviceid_type4 = {
    1 : "NOTIFY_DEVICEID4_CHANGE",
    2 : "NOTIFY_DEVICEID4_DELETE",
}



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Unpack module

Provides the object for managing and unpacking raw data from a working buffer.
"""
import struct
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "2.4"

# Module variables
UNPACK_ERROR = False  # Raise unpack error when True

class Unpack(object):
    """Unpack object

       Usage:
           from packet.unpack import Unpack

           x = Unpack(buffer)

           # Get 32 bytes from the working buffer and move the offset pointer
           data = x.read(32)

           # Get all the unprocessed bytes from the working buffer
           # (all bytes starting from the offset pointer)
           # Do not move the offset pointer
           data = x.getbytes()

           # Get all bytes from the working buffer from the given offset
           # Do not move the offset pointer
           data = x.getbytes(offset)

           # Return the number of unprocessed bytes left in the working buffer
           size = x.size()
           size = len(x)

           # Get the offset pointer
           offset = x.tell()

           # Set the offset pointer
           x.seek(offset)

           # Append the given data to the working buffer
           x.append(data)

           # Insert the given data to the working buffer right before the
           # offset pointer. This resets the working buffer completely
           # and the offset pointer is initialized to zero. It is like
           # re-instantiating the object like:
           #   x = Unpack(data + x.getbytes())
           x.insert(data)

           # Save state
           sid = x.save_state()

           # Restore state
           x.restore_state(sid)

           # Unpack an 'unsigned short' (2 bytes in network order)
           short_int = x.unpack(2, '!H')[0]

           # Unpack different basic types
           char      = x.unpack_char()
           uchar     = x.unpack_uchar()
           short     = x.unpack_short()
           ushort    = x.unpack_ushort()
           int       = x.unpack_int()
           uint      = x.unpack_uint()
           int64     = x.unpack_int64()
           uint64    = x.unpack_uint64()
           data1     = x.unpack_opaque()
           data2     = x.unpack_opaque(64)  # Length of opaque must be <= 64
           data3     = x.unpack_fopaque(32)

           # Get string where length is given as an unsigned integer
           buffer = x.unpack_string()
           # Get string of fixed length
           buffer = x.unpack_string(32)
           # Get string where length is given as a short integer
           buffer = x.unpack_string(Unpack.unpack_short)
           buffer = x.unpack_string(ltype=Unpack.unpack_short)
           # Get string padded to a 4 byte boundary, discard padding bytes
           buffer = x.unpack_string(pad=4)

           # Get an array of unsigned integers
           alist = x.unpack_array()
           # Get a fixed length array of unsigned integers
           alist = x.unpack_array(ltype=10)
           # Get an array of short integers
           alist = x.unpack_array(Unpack.unpack_short)
           # Get an array of strings, the length of the array is given
           # by a short integer
           alist = x.unpack_array(Unpack.unpack_string, Unpack.unpack_short)
           # Get an array of strings, the length of each string is given by
           # a short integer and each string is padded to a 4 byte boundary
           alist = x.unpack_array(Unpack.unpack_string, uargs={'ltype':Unpack.unpack_short, 'pad':4})
           # Get an array of objects decoded by item_obj where the first
           # argument to item_obj is the unpack object, e.g., item = item_obj(x)
           alist = x.unpack_array(item_obj)

           # Get a list of unsigned integers
           alist = x.unpack_list()
           # Get a list of short integers
           alist = x.unpack_list(Unpack.unpack_short)
           # Get a list of strings, the next item flag is given
           # by a short integer
           alist = x.unpack_list(Unpack.unpack_string, Unpack.unpack_short)
           # Get a list of strings, the length of each string is given by
           # a short integer and each string is padded to a 4 byte boundary
           alist = x.unpack_list(Unpack.unpack_string, uargs={'ltype':Unpack.unpack_short, 'pad':4})

           # Unpack a conditional, it unpacks a conditional flag first and
           # if it is true it unpacks the item given and returns it. If the
           # conditional flag decoded is false, the method returns None
           buffer = x.unpack_conditional(Unpack.unpack_opaque)

           # Unpack an array of unsigned integers and convert array into
           # a single long integer
           bitmask = unpack_bitmap()
    """
    def __init__(self, data):
        """Constructor

           Initialize object's private data.

           data:
               Raw packet data
        """
        self._offset = 0
        self._data = data
        self._state = []

    def _get_ltype(self, ltype):
        """Get length of element"""
        if isinstance(ltype, int):
            # An integer is given, just return it
            return ltype
        else:
            # A function is given, return output of function
            return ltype(self)

    def size(self):
        """Return the number of unprocessed bytes left in the working buffer"""
        return len(self._data) - self._offset
    __len__ = size

    def tell(self):
        """Get the offset pointer."""
        return self._offset

    def seek(self, offset):
        """Set the offset pointer."""
        slen = len(self._data)
        if offset > slen:
            offset = slen
        self._offset = offset

    def append(self, data):
        """Append data to the working buffer."""
        self._data += data

    def insert(self, data):
        """Insert data to the beginning of the current working buffer."""
        if len(self._state):
            # Save working buffer in the saved state since the buffer
            # will be overwritten
            state = self._state[-1]
            if len(state) == 2:
                state.append(self._data)
        self._data = data + self._data[self._offset:]
        self._offset = 0

    def save_state(self):
        """Save state and return the state id"""
        sid = len(self._state)
        self._state.append([sid, self._offset])
        return sid

    def restore_state(self, sid):
        """Restore state given by the state id"""
        max = len(self._state)
        while sid < len(self._state):
            state = self._state.pop()
            self._offset = state[1]
            if len(state) == 3:
                self._data = state[2]

    def getbytes(self, offset=None):
        """Get the number of bytes given from the working buffer.
           Do not move the offset pointer.

           offset:
               Starting offset of data to return [default: current offset]
        """
        if offset is None:
            return self._data[self._offset:]
        return self._data[offset:]

    def read(self, size, pad=0):
        """Get the number of bytes given from the working buffer.
           Move the offset pointer.

           size:
               Length of data to get
           pad:
               Get and discard padding bytes [default: 0]
               If given, data is padded to this byte boundary
        """
        buf = self._data[self._offset:self._offset+size]
        if pad > 0:
            # Discard padding bytes
            size += (size+pad-1)/pad*pad - size
        self._offset += size
        dlen = len(self._data)
        if self._offset > dlen:
            self._offset = dlen
        return buf

    def unpack(self, size, fmt):
        """Get the number of bytes given from the working buffer and process
           it according to the given format.
           Return a tuple of unpack items, see struct.unpack.

           size:
               Length of data to process
           fmt:
               Format string on how to process data
        """
        return struct.unpack(fmt, self.read(size))

    def unpack_char(self):
        """Get a signed char"""
        return self.unpack(1, '!b')[0]

    def unpack_uchar(self):
        """Get an unsigned char"""
        return self.unpack(1, '!B')[0]

    def unpack_short(self):
        """Get a signed short integer"""
        return self.unpack(2, '!h')[0]

    def unpack_ushort(self):
        """Get an unsigned short integer"""
        return self.unpack(2, '!H')[0]

    def unpack_int(self):
        """Get a signed integer"""
        return self.unpack(4, '!i')[0]

    def unpack_uint(self):
        """Get an unsigned integer"""
        return self.unpack(4, '!I')[0]

    def unpack_int64(self):
        """Get a signed 64 bit integer"""
        return self.unpack(8, '!q')[0]

    def unpack_uint64(self):
        """Get an unsigned 64 bit integer"""
        return self.unpack(8, '!Q')[0]

    def unpack_opaque(self, maxcount=0):
        """Get a variable length opaque up to a maximum length of maxcount"""
        size = self.unpack_uint()
        if maxcount > 0 and size > maxcount:
            raise Exception, "Opaque exceeds maximum length"
        return self.read(size, pad=4)

    def unpack_fopaque(self, size):
        """Get a fixed length opaque"""
        return self.read(size, pad=4)

    def unpack_string(self, ltype=unpack_uint, pad=0, maxcount=0):
        """Get a variable length string

           ltype:
               Function to decode length of string [default: unpack_uint]
               Could also be given as an integer to have a fixed length string
           pad:
               Get and discard padding bytes [default: 0]
               If given, string is padded to this byte boundary
           maxcount:
               Maximum length of string [default: any length]
        """
        slen = self._get_ltype(ltype)
        if maxcount > 0 and slen > maxcount:
            raise Exception, "String exceeds maximum length"
        return self.read(slen, pad)

    def unpack_array(self, unpack_item=unpack_uint, ltype=unpack_uint, uargs={}, maxcount=0, islist=False):
        """Get a variable length array, the type of objects in the array
           is given by the unpacking function unpack_item and the type
           to decode the length of the array is given by ltype

           unpack_item:
               Unpack function for each item in the array [default: unpack_uint]
           ltype:
               Function to decode length of array [default: unpack_uint]
               Could also be given as an integer to have a fixed length array
           uargs:
               Named arguments to pass to unpack_item function [default: {}]
           maxcount:
               Maximum length of array [default: any length]
        """
        ret = []
        # Get length of array
        slen = self._get_ltype(ltype)
        if maxcount > 0 and slen > maxcount:
            raise Exception, "Array exceeds maximum length"
        while slen > 0:
            try:
                # Unpack each item in the array
                ret.append(unpack_item(self, **uargs))
                if islist:
                    slen = self._get_ltype(ltype)
                else:
                    slen -= 1
            except:
                if UNPACK_ERROR:
                    raise
                break
        return ret

    def unpack_list(self, *kwts, **kwds):
        """Get an indeterminate size list, the type of objects in the list
           is given by the unpacking function unpack_item and the type
           to decode the next item flag is given by ltype

           unpack_item:
               Unpack function for each item in the list [default: unpack_uint]
           ltype:
               Function to decode the next item flag [default: unpack_uint]
           uargs:
               Named arguments to pass to unpack_item function [default: {}]
        """
        kwds['islist'] = True
        return self.unpack_array(*kwts, **kwds)

    def unpack_conditional(self, unpack_item=unpack_uint, ltype=unpack_uint, uargs={}):
        """Get an item if condition flag given by ltype is true, if condition
           flag is false then return None

           unpack_item:
               Unpack function for item if condition is true [default: unpack_uint]
           ltype:
               Function to decode the condition flag [default: unpack_uint]
           uargs:
               Named arguments to pass to unpack_item function [default: {}]
        """
        # Get condition flag
        if self._get_ltype(ltype):
            # Unpack item if condition is true
            return unpack_item(self, **uargs)
        return None

    def unpack_bitmap(self):
        """Unpack an array of unsigned integers and convert array into
           a single long integer
        """
        bitmask = 0
        nshift = 0
        # Unpack array of uint32
        blist = self.unpack_array()
        for bint in blist:
            bitmask += bint << nshift
            nshift += 32
        return bitmask



#===============================================================================
# Copyright 2013 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
RPC constants module

Provide constant values and mapping dictionaries for the RPC layer.
"""
import nfstest_config as c

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2013 NetApp, Inc."
__license__   = "GPL v2"

# msg_type
CALL  = 0
REPLY = 1
msg_type = {
    0: 'CALL',
    1: 'REPLY',
}

# reply_stat
MSG_ACCEPTED = 0
MSG_DENIED   = 1
reply_stat = {
    0: 'MSG_ACCEPTED',
    1: 'MSG_DENIED',
}

# accept_stat
SUCCESS       = 0  # RPC executed successfully
PROG_UNAVAIL  = 1  # remote hasn't exported program
PROG_MISMATCH = 2  # remote can't support version #
PROC_UNAVAIL  = 3  # program can't support procedure
GARBAGE_ARGS  = 4  # procedure can't decode params
SYSTEM_ERR    = 5  # e.g. memory allocation failure
accept_stat = {
    0: 'SUCCESS',
    1: 'PROG_UNAVAIL',
    2: 'PROG_MISMATCH',
    3: 'PROC_UNAVAIL',
    4: 'GARBAGE_ARGS',
    5: 'SYSTEM_ERR',
}

# reject_stat
RPC_MISMATCH = 0, # RPC version number != 2
AUTH_ERROR   = 1  # remote can't authenticate caller
reject_stat = {
    0: 'RPC_MISMATCH',
    1: 'AUTH_ERROR',
}

# auth_stat
AUTH_OK                 = 0  # success
                             # failed at remote end
AUTH_BADCRED            = 1  # bad credential (seal broken)
AUTH_REJECTEDCRED       = 2  # client must begin new session
AUTH_BADVERF            = 3  # bad verifier (seal broken)
AUTH_REJECTEDVERF       = 4  # verifier expired or replayed
AUTH_TOOWEAK            = 5  # rejected for security reasons
                             # failed locally
AUTH_INVALIDRESP        = 6  # bogus response verifier
AUTH_FAILED             = 7  # reason unknown
                             # AUTH_KERB errors; deprecated.  See [RFC2695]
AUTH_KERB_GENERIC       = 8  # kerberos generic error
AUTH_TIMEEXPIRE         = 9  # time of credential expired
AUTH_TKT_FILE          = 10  # problem with ticket file
AUTH_DECODE            = 11  # can't decode authenticator
AUTH_NET_ADDR          = 12  # wrong net address in ticket
                             # RPCSEC_GSS GSS related errors
RPCSEC_GSS_CREDPROBLEM = 13  # no credentials for user
RPCSEC_GSS_CTXPROBLEM  = 14  # problem with context
auth_stat = {
     0: 'AUTH_OK',
     1: 'AUTH_BADCRED',
     2: 'AUTH_REJECTEDCRED',
     3: 'AUTH_BADVERF',
     4: 'AUTH_REJECTEDVERF',
     5: 'AUTH_TOOWEAK',
     6: 'AUTH_INVALIDRESP',
     7: 'AUTH_FAILED',
     8: 'AUTH_KERB_GENERIC',
     9: 'AUTH_TIMEEXPIRE',
    10: 'AUTH_TKT_FILE',
    11: 'AUTH_DECODE',
    12: 'AUTH_NET_ADDR',
    13: 'RPCSEC_GSS_CREDPROBLEM',
    14: 'RPCSEC_GSS_CTXPROBLEM',
}

# authentication flavor numbers
AUTH_NONE  = 0  # no authentication, see RFC 1831
                # a.k.a. AUTH_NULL
AUTH_SYS   = 1  # unix style (uid+gids), RFC 1831
                # a.k.a. AUTH_UNIX
AUTH_SHORT = 2  # short hand unix style, RFC 1831
AUTH_DH    = 3  # des style (encrypted timestamp)
                # a.k.a. AUTH_DES, see RFC 2695
AUTH_KERB  = 4  # kerberos auth, see RFC 2695
AUTH_RSA   = 5  # RSA authentication
RPCSEC_GSS = 6  # GSS-based RPC security for auth,
                # integrity and privacy, RPC 5403
auth_flavor = {
    0: 'AUTH_NONE',
    1: 'AUTH_SYS',
    2: 'AUTH_SHORT',
    3: 'AUTH_DH',
    4: 'AUTH_KERB',
    5: 'AUTH_RSA',
    6: 'RPCSEC_GSS',
}



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from mount3.x on Sun Oct 04 08:09:27 2015
"""
MOUNTv3 decoding module
"""
import nfstest_config as c
from packet.utils import *
from baseobj import BaseObj
import mount3_const as const
from packet.unpack import Unpack

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "3.0"

# Basic Data Types
fhandle3 = lambda unpack: StrHex(unpack.unpack_opaque(const.FHSIZE3))
dirpath3 = lambda unpack: unpack.unpack_opaque(const.MNTPATHLEN)
name3    = lambda unpack: unpack.unpack_opaque(const.MNTNAMLEN)

class mountstat3(Enum):
    """enum mountstat3"""
    _enumdict = const.mountstat3

class rpc_auth_flavors(Enum):
    """enum rpc_auth_flavors"""
    _enumdict = const.rpc_auth_flavors

# MNT3res MOUNTPROC3_MNT(dirpath3) = 1;
class MNT3args(BaseObj):
    """
       struct MNT3args {
           dirpath3 path;
       };
    """
    # Class attributes
    _strfmt1  = "{0}"
    _attrlist = ("path",)

    def __init__(self, unpack):
        self.path = dirpath3(unpack)

class MNT3resok(BaseObj):
    """
       struct MNT3resok {
           fhandle3         fh;
           rpc_auth_flavors auth_flavors<>;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32} auth_flavors:{1}"
    _attrlist = ("fh", "auth_flavors")

    def __init__(self, unpack):
        self.fh           = fhandle3(unpack)
        self.auth_flavors = unpack.unpack_array(rpc_auth_flavors)

class MNT3res(BaseObj):
    """
       union switch MNT3res (mountstat3 status) {
           case const.MNT3_OK:
               MNT3resok mountinfo;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", mountstat3(unpack))
        if self.status == const.MNT3_OK:
            self.set_attr("mountinfo", MNT3resok(unpack), switch=True)

# mountlist MOUNTPROC3_DUMP(void) = 2;
class mountentry(BaseObj):
    """
       struct mountentry {
           name3      hostname;
           dirpath3   directory;
           mountentry *next;
       };
    """
    # Class attributes
    _attrlist = ("hostname", "directory")

    def __init__(self, unpack):
        self.hostname  = name3(unpack)
        self.directory = dirpath3(unpack)

class DUMP3res(BaseObj):
    """
       struct DUMP3res {
           mountentry *mountlist;
       };
    """
    # Class attributes
    _strfmt1  = ""
    _attrlist = ("mountlist",)

    def __init__(self, unpack):
        self.mountlist = unpack.unpack_list(mountentry)

# void MOUNTPROC3_UMNT(dirpath3) = 3;
class UMNT3args(BaseObj):
    """
       struct UMNT3args {
           dirpath3 path;
       };
    """
    # Class attributes
    _strfmt1  = "{0}"
    _attrlist = ("path",)

    def __init__(self, unpack):
        self.path = dirpath3(unpack)

# void MOUNTPROC3_UMNTALL(void) = 4;
#
# EXPORT3res MOUNTPROC3_EXPORT(void) = 5;
class exportnode3(BaseObj):
    """
       struct exportnode3 {
           dirpath3    dir;
           name3       *groups;
           exportnode3 *next;
       };
    """
    # Class attributes
    _attrlist = ("dir", "groups")

    def __init__(self, unpack):
        self.dir    = dirpath3(unpack)
        self.groups = unpack.unpack_list(name3)

class EXPORT3res(BaseObj):
    """
       struct EXPORT3res {
           exportnode3 *exports;
       };
    """
    # Class attributes
    _strfmt1  = ""
    _attrlist = ("exports",)

    def __init__(self, unpack):
        self.exports = unpack.unpack_list(exportnode3)

# Procedures
class mount_proc3(Enum):
    """enum mount_proc3"""
    _enumdict = const.mount_proc3

# Version 3 of the mount protocol used with
# version 3 of the NFS protocol.
class MOUNT3args(RPCload):
    """
       union switch MOUNT3args (mount_proc3 procedure) {
           case const.MOUNTPROC3_NULL:
               void;
           case const.MOUNTPROC3_MNT:
               MNT3args opmnt;
           case const.MOUNTPROC3_DUMP:
               void;
           case const.MOUNTPROC3_UMNT:
               UMNT3args opumnt;
           case const.MOUNTPROC3_UMNTALL:
               void;
           case const.MOUNTPROC3_EXPORT:
               void;
       };
    """
    # Class attributes
    _pindex  = 11
    _strname = "MOUNT"

    def __init__(self, unpack, procedure):
        self.set_attr("procedure", mount_proc3(procedure))
        if self.procedure == const.MOUNTPROC3_NULL:
            self.set_strfmt(2, "NULL()")
        elif self.procedure == const.MOUNTPROC3_MNT:
            self.set_attr("opmnt", MNT3args(unpack), switch=True)
        elif self.procedure == const.MOUNTPROC3_DUMP:
            self.set_strfmt(2, "DUMP3args()")
        elif self.procedure == const.MOUNTPROC3_UMNT:
            self.set_attr("opumnt", UMNT3args(unpack), switch=True)
        elif self.procedure == const.MOUNTPROC3_UMNTALL:
            self.set_strfmt(2, "UMNTALL3args()")
        elif self.procedure == const.MOUNTPROC3_EXPORT:
            self.set_strfmt(2, "EXPORT3args()")
        self.argop = self.procedure
        self.op    = self.procedure

class MOUNT3res(RPCload):
    """
       union switch MOUNT3res (mount_proc3 procedure) {
           case const.MOUNTPROC3_NULL:
               void;
           case const.MOUNTPROC3_MNT:
               MNT3res opmnt;
           case const.MOUNTPROC3_DUMP:
               DUMP3res opdump;
           case const.MOUNTPROC3_UMNT:
               void;
           case const.MOUNTPROC3_UMNTALL:
               void;
           case const.MOUNTPROC3_EXPORT:
               EXPORT3res opexport;
       };
    """
    # Class attributes
    _pindex  = 11
    _strname = "MOUNT"

    def __init__(self, unpack, procedure):
        self.set_attr("procedure", mount_proc3(procedure))
        if self.procedure == const.MOUNTPROC3_NULL:
            self.set_strfmt(2, "NULL()")
        elif self.procedure == const.MOUNTPROC3_MNT:
            self.set_attr("opmnt", MNT3res(unpack), switch=True)
        elif self.procedure == const.MOUNTPROC3_DUMP:
            self.set_attr("opdump", DUMP3res(unpack), switch=True)
        elif self.procedure == const.MOUNTPROC3_UMNT:
            self.set_strfmt(2, "UMNT3res()")
        elif self.procedure == const.MOUNTPROC3_UMNTALL:
            self.set_strfmt(2, "UMNTALL3res()")
        elif self.procedure == const.MOUNTPROC3_EXPORT:
            self.set_attr("opexport", EXPORT3res(unpack), switch=True)
        self.resop = self.procedure
        self.op    = self.procedure



#===============================================================================
# Copyright 2013 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Remote procedure module

Provides a set of tools for executing a wide range commands, statements,
expressions or functions on a remote host by running a server process
on the remote host serving requests without disconnecting. This allows
for a sequence of operations to be done remotely and not losing state.
A file could be opened remotely, do some other things and then write
to the same opened file without opening the file again.

The remote server can be executed as a different user by using the sudo
option and sending seteuid. The server can be executed locally as well
using fork when running as the same user or using the shell when the
sudo option is used.

In order to use this module the user id must be able to 'ssh' to the
remote host without the need for a password.
"""
import os
import time
import types
import inspect
import nfstest_config as c
from baseobj import BaseObj
from subprocess import Popen, PIPE
from multiprocessing.connection import Client, Listener

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2013 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.2"

# Constants
PORT = 9900

# Imports needed for RemoteServer class
IMPORTS = """
import time
import types
from multiprocessing.connection import Listener
"""

class RemoteServer:
    def __init__(self, port, logfile=None):
        """Remote procedure server"""
        self.port     = port
        self.logfile  = logfile
        self.fd       = None
        self.conn     = None
        self.listener = None

    def __del__(self):
        """Destructor"""
        if self.fd is not None:
            self.fd.close()
        if self.listener is not None:
            self.listener.close()

    def log(self, msg):
        """Write message to log file"""
        if self.fd is None:
            return
        curtime = time.time()
        msec = "%06d" % (1000000 * (curtime - int(curtime)))
        tstamp = time.strftime("%H:%M:%S.", time.localtime(curtime)) + msec
        self.fd.write(tstamp + " - " + msg + "\n")
        self.fd.flush()

    def start(self):
        if self.logfile is not None:
            self.fd = open(self.logfile, "w", 0)
        self.listener = Listener(("", self.port))
        self.conn = self.listener.accept()
        self.log("Connection accepted\n")

        while True:
            msg = self.conn.recv()
            self.log("RECEIVED: %r" % msg)
            if type(msg) is dict:
                try:
                    # Get command
                    cmd  = msg.get("cmd")
                    # Get function/statement/expression and positional arguments
                    kwts = msg.get("kwts", ())
                    fstr = kwts[0]
                    kwts = kwts[1:]
                    # Get named arguments
                    kwds = msg.get("kwds", {})

                    if cmd == "run":
                        # Find if function is defined
                        if type(fstr) in [types.FunctionType, types.BuiltinFunctionType, types.MethodType]:
                            # This is a function
                            func = fstr
                        else:
                            # Find symbol in locals then in globals
                            func = locals().get(fstr)
                            if func is None:
                                func = globals().get(fstr)
                        if func is None:
                            raise Exception("function not found")
                        # Run function with all its arguments
                        out = func(*kwts, **kwds)
                        self.log("RESULT: " + repr(out))
                        self.conn.send(out)
                    elif cmd == "eval":
                        # Evaluate expression
                        out = eval(fstr)
                        self.log("RESULT: " + repr(out))
                        self.conn.send(out)
                    elif cmd == "exec":
                        # Execute statement
                        exec(fstr)
                        self.log("EXEC done")
                        self.conn.send(None)
                    else:
                        emsg = "Unknown procedure"
                        self.log("ERROR: %s" % emsg)
                        self.conn.send(Exception(emsg))
                except Exception as e:
                    self.log("ERROR: %r" % e)
                    self.conn.send(e)
            if msg == "close":
                # Request to close the connection,
                # exit the loop and terminate the server
                self.conn.close()
                break

class Rexec(BaseObj):
    """Rexec object

       Rexec() -> New remote procedure object

       Arguments:
           servername:
               Name or IP address of remote server
           logfile:
               Name of logfile to create on remote server
           sudo:
               Run remote server as root

       Usage:
           from nfstest.rexec import Rexec

           # Function to be defined at remote host
           def add_one(n):
               return n + 1

           # Function to be defined at remote host
           def get_time(delay=0):
               time.sleep(delay)
               return time.time()

           # Create remote procedure object
           x = Rexec("192.168.0.85")

           # Define function at remote host
           x.rcode(add_one)

           # Evaluate the expression calling add_one()
           out = x.reval("add_one(67)")

           # Run the function with the given argument
           out = x.run("add_one", 7)

           # Run built-in functions
           import time
           out = x.run(time.time)

           # Import libraries and symbols
           x.rimport("time", ["sleep"])
           x.run("sleep", 2)

           # Define function at remote host -- since function uses the
           # time module, this module must be first imported
           x.rimport("time")
           x.rcode(get_time)

           # Evaluate the expression calling get_time()
           out = x.reval("get_time()")

           # Run the function with the given argument
           out = x.run("get_time", 10)

           # Open file on remote host
           fd = x.run(os.open, "/tmp/testfile", os.O_WRONLY|os.O_CREAT|os.O_TRUNC)
           count = x.run(os.write, fd, "hello there\n")
           x.run(os.close, fd)

           # Use of positional arguments
           out = x.run("get_time", 2)

           # Use of named arguments
           out = x.run("get_time", delay=2)

           # Use of NOWAIT option for long running functions so other things
           # can be done while waiting
           x.run("get_time", 2, NOWAIT=True)
           while True:
               # Poll every 0.1 secs to see if function has finished
               if x.poll(0.1):
                   # Get results
                   out = x.results()
                   break

           # Create remote procedure object as a different user
           # First, run the remote server as root
           x = Rexec("192.168.0.85", sudo=True)
           # Then set the effective user id
           x.run(os.seteuid, 1000)
    """
    def __init__(self, servername=None, logfile=None, sudo=False, sync_timeout=0.1):
        """Constructor

           Initialize object's private data.

           servername:
               Host name or IP address of host where remote server will run
               [Default: None (run locally)]
           logfile:
               Pathname of log file to be created on remote host
               [Default: None]
           sudo:
               Run remote procedure server as root
               [Default: False]
           sync_timeout:
               Timeout used for synchronizing the connection stream
               [Default: 0.1]
        """
        global PORT
        self.pid     = None
        self.conn    = None
        self.process = None
        self.remote  = False
        self.servername   = servername
        self.logfile      = logfile
        self.sudo         = sudo
        self.sync_timeout = sync_timeout
        if os.getuid() == 0:
            # Already running as root
            self.sudo = True
            sudo = False

        if not sudo and servername in [None, "", "localhost", "127.0.0.1"]:
            # Start remote server locally via fork when sudo is not set
            servername = ""
            self.pid = os.fork()
            if self.pid == 0:
                # This is the child process
                RemoteServer(PORT, self.logfile).start()
                os._exit(0)
        else:
            # Start server on remote host or locally if sudo is set
            server_code  = IMPORTS
            server_code += "".join(inspect.getsourcelines(RemoteServer)[0])
            server_code += "RemoteServer(%d, %r).start()\n" % (PORT, self.logfile)
            # Execute minimal python script to execute the source code
            # given in standard input
            pysrc = "import sys; exec(sys.stdin.read(%d))" % len(server_code)
            cmdlist = ["python", "-c", repr(pysrc)]

            if sudo:
                cmdlist.insert(0, "sudo")

            if servername not in [None, "", "localhost", "127.0.0.1"]:
                # Run remote process via ssh
                cmdlist = ["ssh", servername] + cmdlist
                self.process = Popen(cmdlist, shell=False, stdin=PIPE)
                self.remote = True
            else:
                # Run local process via the shell
                servername = ""
                self.process = Popen(" ".join(cmdlist), shell=True, stdin=PIPE)

            # Send the server code to be executed via standard input
            self.process.stdin.write(server_code)

        # Connect to remote server
        address = (servername, PORT)
        self.conn = Client(address)
        PORT += 1

    def __del__(self):
        """Destructor"""
        self.close()

    def close(self):
        """Close connection to remote server"""
        if self.conn:
            # Send command to exit main loop
            self.conn.send("close")
            self.conn.close()
            self.conn = None
        # Wait for remote server to finish
        if self.pid:
            os.waitpid(self.pid, 0)
            self.pid = None
        elif self.process:
            self.process.wait()
            self.process = None

    def _send_cmd(self, cmd, *kwts, **kwds):
        """Internal method to send commands to remote server"""
        nowait = kwds.pop("NOWAIT", False)
        self.conn.send({"cmd": cmd, "kwts": kwts, "kwds": kwds})
        if nowait:
            # NOWAIT option is specified, so return immediately
            # Use poll() method to check if any data is available
            # Use results() method to get pending results from function
            return
        return self.results()

    def wait(self, objlist=None, timeout=0):
        """Return a list of Rexec objects where data is available to be read

           objlist:
               List of Rexec objects to poll, if not given use current object
           timeout:
               Maximum time in seconds to block, if timeout is None then
               an infinite timeout is used
        """
        ret = []
        if objlist is None:
            # Use current object as default
            objlist = [self]

        for obj in objlist:
            if obj.poll(timeout):
                ret.append(obj)
            # Just check all other objects if they are ready now
            timeout = 0
        return ret if len(ret) else None

    def poll(self, timeout=0):
        """Return whether there is any data available to be read

           timeout:
               Maximum time in seconds to block, if timeout is None then
               an infinite timeout is used
        """
        return self.conn.poll(timeout)

    def results(self):
        """Return pending results"""
        while True:
            out = self.conn.recv()
            if isinstance(out, Exception):
                raise out
            elif out is None and self.poll(self.sync_timeout):
                # Try to re-sync when recv() returns None and there is
                # still data in the buffer
                continue
            return out

    def rexec(self, expr):
        """Execute statement on remote server"""
        return self._send_cmd("exec", expr)

    def reval(self, expr):
        """Evaluate expression on remote server"""
        return self._send_cmd("eval", expr)

    def run(self, *kwts, **kwds):
        """Run function on remote server

           The first positional argument is the function to be executed.
           All other positional arguments and any named arguments are treated
           as arguments to the function
        """
        return self._send_cmd("run", *kwts, **kwds)

    def rcode(self, code):
        """Define function on remote server"""
        codesrc = "".join(inspect.getsourcelines(code)[0])
        self.rexec(codesrc)

    def rimport(self, module, symbols=[]):
        """Import module on remote server

           module:
               Module to import in the remote server
           symbols:
               If given, import only these symbols from the module
        """
        # Import module
        if len(symbols) == 0:
            self.rexec("import %s" % module)
            symbols = [module]
        else:
            self.rexec("from %s import %s" % (module, ",".join(symbols)))
        # Make all symbols global
        for item in symbols:
            self.rexec("globals()['%s']=locals()['%s']" % (item, item))



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
File I/O module

Provides an interface to create and manipulate files of different types.
The arguments allow running for a specified period of time as well as running
multiple processes. Each process modifies a single file at a time and the
file name space is different for each process so there are no collisions
between two different processes modifying the same file.

File types:
  - Regular file
  - Hard link
  - Symbolic link

File operations:
  - Open (create or re-open)
  - Open downgrade
    This is done by opening the file for read and write, then the file is
    opened again as read only and finally closing the read and write file
    descriptor
  - Read (sequential or random access)
  - Write (sequential or random access)
  - Remove
  - Rename
  - Truncate (path or file descriptor)
  - Readdir
  - Lock
  - Unlock
  - Tlock
"""
import os
import re
import sys
import time
import errno
import fcntl
import ctypes
import signal
import struct
import formatstr
import traceback
import subprocess
from random import Random
import nfstest_config as c
from baseobj import BaseObj
from formatstr import str_units, int_units
from multiprocessing import Process,JoinableQueue

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.3"

# Default values
P_SEED       = None
P_NPROCS     = 1
P_RUNTIME    = 0
P_VERBOSE    = "none"
P_CREATELOG  = False
P_CREATELOGS = False
P_CREATE     = 5.0
P_OSYNC      = 10.0
P_FSYNC      = 2.0
P_READ       = 40.0
P_WRITE      = 40.0
P_RDWR       = 20.0
P_ODGRADE    = 5.0
P_RANDIO     = 50.0
P_RDWRONLY   = False
P_DIRECT     = False
P_TMPDIR     = "/tmp"
P_IODELAY    = 0.0

P_RENAME     = 5.0
P_REMOVE     = 5.0
P_TRUNC      = 2.0
P_FTRUNC     = 2.0
P_LINK       = 1.0
P_SLINK      = 0.2
P_READDIR    = 0.5
P_LOCK       = 20.0
P_UNLOCK     = 80.0
P_TLOCK      = 20.0
P_LOCKFULL   = 50.0

P_FILESIZE   = "1m"
P_FSIZEDEV   = "256k"
P_RSIZE      = "64k"
P_WSIZE      = "64k"
P_RSIZEDEV   = "8k"
P_WSIZEDEV   = "8k"
P_SIZEMULT   = "1.0"

# Minimum number of files to create before doing any other
# file operations like remove, rename, etc.
MIN_FILES = 10

# Mapping dictionaries
LOCKMAP = {
    fcntl.F_RDLCK: "RDLCK",
    fcntl.F_WRLCK: "WRLCK",
    fcntl.F_UNLCK: "UNLCK",
}

OPENMAP = {
    os.O_RDONLY: "O_RDONLY",
    os.O_WRONLY: "O_WRONLY",
    os.O_RDWR:   "O_RDWR",
    os.O_CREAT:  "O_CREAT",
    os.O_TRUNC:  "O_TRUNC",
    os.O_SYNC:   "O_SYNC",
}

class TermSignal(Exception):
    """Exception to be raised on SIGTERM signal"""
    pass

def stop_handler(signum, frame):
    """Signal handler to catch SIGTERM and allow for graceful termination
       of subprocesses
    """
    raise TermSignal("Terminating process!")

# File object
class FileObj(BaseObj): pass

class FileIO(BaseObj):
    """FileIO object

       Usage:
           from nfstest.file_io import FileIO

           # Instantiate FileIO object given top level directory
           x = FileIO(datadir="/tmp/data")

           # Run workload creating the top level directory if necessary
           x.run()
    """
    def __init__(self, **kwargs):
        """Constructor

           Initialize object's private data

           datadir:
               Top level directory where files will be created,
               it will be created if it does not exist
           seed:
               Seed to initialized the random number generator
               [default: automatically generated]
           nprocs:
               Number of processes to use [default: 1]
           runtime:
               Run time [default: 0 (indefinitely)]
           verbose:
               Verbose level: none|info|debug|dbg1-7|all [default: 'none']
           exiterr:
               Exit on first error [default: False]
           read:
               Read file percentage [default: 40.0]
           write:
               Write file percentage [default: 40.0]
           rdwr:
               Read/write file percentage [default: 20.0]
           randio:
               Random file access percentage [default: 50.0]
           iodelay:
               Seconds to delay I/O operations [default: 0.0]
           direct:
               Use direct I/O [default: False]
           rdwronly:
               Use read and write only, no rename, remove, etc. [default: False]
           create:
               Create file percentage [default: 5.0]
           odgrade:
               Open downgrade percentage [default: 5.0]
           osync:
               Open file with O_SYNC [default: 10.0]
           fsync:
               Percentage of fsync after write [default: 2.0]
           rename:
               Rename file percentage [default: 5.0]
           remove:
               Remove file percentage [default: 5.0]
           trunc:
               Truncate file percentage [default: 2.0]
           ftrunc:
               Truncate opened file percentage [default: 2.0]
           link:
               Create hard link percentage [default: 1.0]
           slink:
               Create symbolic link percentage [default: 0.2]
           readdir:
               List contents of directory percentage [default: 0.5]
           lock:
               Lock file percentage [default: 20.0]
           unlock:
               Unlock file percentage [default: 80.0]
           tlock:
               Lock test percentage [default: 20.0]
           lockfull:
               Lock full file percentage [default: 50.0]
           minfiles:
               Minimum number of files to create before any file operation
               is executed [default: 10]
           fsizeavg:
               File size average [default: 1m]
           fsizedev:
               File size standard deviation [default: 256k]
           rsize:
               Read block size [default: 64k]
           rsizedev:
               Read block size standard deviation [default: 8k]
           wsize:
               Write block size [default: 64k]
           wsizedev:
               Write block size standard deviation [default: 8k]
           sizemult:
               Size multiplier [default: 1.0]
           createlog:
               Create log file [default: False]
           createlogs:
               Create a log file for each process [default: False]
           logdir:
               Log directory [default: '/tmp']
        """
        self.progname   = os.path.basename(sys.argv[0])
        self.datadir    = kwargs.pop("datadir",    None)
        self.seed       = kwargs.pop("seed",       P_SEED)
        self.nprocs     = kwargs.pop("nprocs",     P_NPROCS)
        self.runtime    = kwargs.pop("runtime",    P_RUNTIME)
        self.verbose    = kwargs.pop("verbose",    P_VERBOSE)
        self.createlog  = kwargs.pop("createlog",  P_CREATELOG)
        self.createlogs = kwargs.pop("createlogs", P_CREATELOGS)
        self.create     = kwargs.pop("create",     P_CREATE)
        self.osync      = kwargs.pop("osync",      P_OSYNC)
        self.fsync      = kwargs.pop("fsync",      P_FSYNC)
        self.read       = kwargs.pop("read",       None)
        self.write      = kwargs.pop("write",      None)
        self.rdwr       = kwargs.pop("rdwr",       None)
        self.odgrade    = kwargs.pop("odgrade",    P_ODGRADE)
        self.randio     = kwargs.pop("randio",     P_RANDIO)
        self.rdwronly   = kwargs.pop("rdwronly",   P_RDWRONLY)
        self.iodelay    = kwargs.pop("iodelay",    P_IODELAY)
        self.direct     = kwargs.pop("direct",     P_DIRECT)
        self.logdir     = kwargs.pop("logdir",     P_TMPDIR)
        self.exiterr    = kwargs.pop("exiterr",    False)
        self.minfiles   = kwargs.pop("minfiles",   str(MIN_FILES))

        if self.datadir is None:
            print "Error: datadir is required"
            sys.exit(2)

        data = [int(x) for x in self.minfiles.split(",")]
        if len(data) == 1:
            self.up_minfiles = -1
            self.top_minfiles  = data[0]
            self.bot_minfiles  = data[0]
        elif len(data) > 1:
            self.up_minfiles = 0
            self.top_minfiles  = max(data)
            self.bot_minfiles  = min(data)
        else:
            print "Error: option minfiles must be an integer or two integers separated by a ',': %s" % self.minfiles
            sys.exit(2)
        self.minfiles = self.top_minfiles

        if self.rdwronly:
            # When rdwronly option is given, set all options for manipulating
            # files to zero if not explicitly given
            self.rename   = kwargs.pop("rename",   0)
            self.remove   = kwargs.pop("remove",   0)
            self.trunc    = kwargs.pop("trunc",    0)
            self.ftrunc   = kwargs.pop("ftrunc",   0)
            self.link     = kwargs.pop("link",     0)
            self.slink    = kwargs.pop("slink",    0)
            self.readdir  = kwargs.pop("readdir",  0)
            self.lock     = kwargs.pop("lock",     0)
            self.unlock   = kwargs.pop("unlock",   0)
            self.tlock    = kwargs.pop("tlock",    0)
            self.lockfull = kwargs.pop("lockfull", 0)
        else:
            self.rename   = kwargs.pop("rename",   P_RENAME)
            self.remove   = kwargs.pop("remove",   P_REMOVE)
            self.trunc    = kwargs.pop("trunc",    P_TRUNC)
            self.ftrunc   = kwargs.pop("ftrunc",   P_FTRUNC)
            self.link     = kwargs.pop("link",     P_LINK)
            self.slink    = kwargs.pop("slink",    P_SLINK)
            self.readdir  = kwargs.pop("readdir",  P_READDIR)
            self.lock     = kwargs.pop("lock",     P_LOCK)
            self.unlock   = kwargs.pop("unlock",   P_UNLOCK)
            self.tlock    = kwargs.pop("tlock",    P_TLOCK)
            self.lockfull = kwargs.pop("lockfull", P_LOCKFULL)

        # Get size multiplier
        sizemult = kwargs.pop("sizemult", P_SIZEMULT)
        if re.search("^[\d\.]+$", sizemult):
            self.sizemult = float(sizemult)
        else:
            self.sizemult = float(int_units(sizemult))
        # Convert sizes and apply multiplier
        self.fsizeavg  = int(self.sizemult * int_units(kwargs.pop("fsizeavg", P_FILESIZE)))
        self.fsizedev  = int(self.sizemult * int_units(kwargs.pop("fsizedev", P_FSIZEDEV)))
        self.rsize     = int(self.sizemult * int_units(kwargs.pop("rsize",    P_RSIZE)))
        self.wsize     = int(self.sizemult * int_units(kwargs.pop("wsize",    P_WSIZE)))
        self.rsizedev  = int(self.sizemult * int_units(kwargs.pop("rsizedev", P_RSIZEDEV)))
        self.wsizedev  = int(self.sizemult * int_units(kwargs.pop("wsizedev", P_WSIZEDEV)))

        if self.direct:
            # When using direct I/O, use fixed read/write block sizes
            self.rsizedev = 0
            self.wsizedev = 0

        # Initialize counters
        self.rbytes   = 0
        self.wbytes   = 0
        self.nopen    = 0
        self.nopendgr = 0
        self.nosync   = 0
        self.nclose   = 0
        self.nread    = 0
        self.nwrite   = 0
        self.nfsync   = 0
        self.nrename  = 0
        self.nremove  = 0
        self.ntrunc   = 0
        self.nftrunc  = 0
        self.nlink    = 0
        self.nslink   = 0
        self.nreaddir = 0
        self.nlock    = 0
        self.nunlock  = 0
        self.ntlock   = 0
        self.stime    = 0

        # Set read and write option percentages
        total = 100.0
        if self.rdwr is None:
            if self.read is None and self.write is None:
                # None of the read and write options are given, use defaults
                self.read  = P_READ
                self.write = P_WRITE
                self.rdwr  = P_RDWR
            elif self.read is None or self.write is None:
                # If only read or write is given, don't use rdwr
                self.rdwr = 0.0
            else:
                # If both read and write are given, set rdwr to add up to 100
                self.rdwr = max(0.0, total - self.read - self.write)
        else:
            # Option rdwr is given, calculate remainder left for read and write
            total -= self.rdwr

        if self.read is None and self.write is None:
            # Only rdwr is given, distribute remainder equally
            # between read and write
            self.read = total/2.0
            self.write = total - self.read
        elif self.read is None and self.write is not None:
            # Option rdwr and write are given, set read percentage
            self.read = total - self.write
        elif self.read is not None and self.write is None:
            # Option rdwr and read are given, set write percentage
            self.write = total - self.read

        # Verify read and write options add up to 100 percent
        total = abs(self.read) + abs(self.write) + abs(self.rdwr)
        if total != 100.0:
            print "Total for read, write and rdwr must be == 100"
            sys.exit(2)

        # Set verbose level mask
        self.debug_level(self.verbose)

        # Set timestamp format to include the date and time
        self.tstamp(fmt="{0:date:%Y-%m-%d %H:%M:%S.%q}  ")

        self.logbase = None
        if self.createlog or self.createlogs:
            # Create main log file
            datetimestr = self.timestamp("{0:date:%Y%m%d%H%M%S_%q}")
            logname = "%s_%s" % (self.progname, datetimestr)
            self.logbase = os.path.join(self.logdir, logname)
            self.logfile = self.logbase + ".log"
            self.open_log(self.logfile)

        # Multiprocessing
        self.tid   = 0
        self.queue = None

        # Memory buffers
        self.fbuffers = []
        self.PAGESIZE = os.sysconf(os.sysconf_names['SC_PAGESIZE'])

        # Load share library for calling C library functions
        try:
            # Linux
            self.libc = ctypes.CDLL('libc.so.6', use_errno=True)
        except:
            # MacOS
            self.libc = ctypes.CDLL('libc.dylib', use_errno=True)
        self.libc.malloc.argtypes = [ctypes.c_long]
        self.libc.malloc.restype = ctypes.c_void_p
        self.libc.posix_memalign.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_long, ctypes.c_long]
        self.libc.posix_memalign.restype = ctypes.c_int
        self.libc.read.argtypes = [ctypes.c_int, ctypes.c_void_p, ctypes.c_long]
        self.libc.read.restype = ctypes.c_int
        self.libc.write.argtypes = [ctypes.c_int, ctypes.c_void_p, ctypes.c_long]
        self.libc.write.restype = ctypes.c_int
        self.libc.lseek.argtypes = [ctypes.c_int, ctypes.c_long, ctypes.c_int]
        self.libc.lseek.restype = ctypes.c_long
        self.libc.memcpy.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_long]
        self.libc.memcpy.restype = ctypes.c_void_p

    def __del__(self):
        """Destructor"""
        if getattr(self, 'logfile', None):
            print "\nLogfile: %s" % self.logfile

    def _dprint(self, level, msg):
        """Local dprint function, if called from a subprocess send the
           message to the main process, otherwise use dprint on message
        """
        if self.queue and not self.createlogs:
            # Send message to main process
            self.queue.put([level,msg])
        else:
            # Display message and send it to the log file
            self.dprint(level, msg)

    def _get_tree(self):
        """Read top level directory for existing files to populate database
           This is used so it can be run in the same top level directory
           multiple times
        """
        for entry in os.listdir(self.datadir):
            # Must match file names given by _newname
            if not re.search(r'^f[\dA-F]+$', entry):
                continue
            # Get tid from file name
            tid = int(entry[1:self.bidx], 16)
            if self.tid != tid:
                continue
            # Get index from file name and set it
            index = int(entry[self.bidx:], 16)
            if self.n_index <= index:
                self.n_index = index + 1

            # Get file size and append it to database
            absfile = os.path.join(self.datadir, entry)
            try:
                fst = os.stat(absfile)
                size = fst.st_size
            except:
                size = 0
            fileobj = FileObj(name=entry, size=size)
            fileobj.debug_repr(1)
            if os.path.islink(absfile):
                fileobj.srcname = os.path.basename(os.readlink(absfile))
            self.n_files.append(fileobj)

    def _newname(self):
        """Create new file name"""
        name = "%s%06X" % (self.basename, self.n_index)
        self.n_index += 1
        return name

    def _percent(self, pvalue):
        """Test percent value"""
        if pvalue >= 100:
            return True
        elif pvalue <= 0:
            return False
        return self.random.randint(0,9999) < 100*pvalue

    def _get_fileobj(self):
        """Get a random file object"""
        # Number of files available
        nlen = len(self.n_files)
        self.findex = self.random.randint(0, nlen-1)
        return self.n_files[self.findex]

    def _getiolist(self, size, iswrite):
        """Return list of I/O blocks to read/write"""
        iolist = []
        if iswrite:
            bsize = self.wsize
            bdev  = self.wsizedev
        else:
            bsize = self.rsize
            bdev  = self.rsizedev

        tsize = 0
        offset = 0
        while tsize < size:
            block = {}
            if self.direct:
                # Direct I/O uses same block size for all blocks
                blocksize = bsize
            else:
                # Buffered I/O uses different block sizes
                blocksize = int(abs(self.random.gauss(bsize, bdev)))
            if tsize + blocksize > size:
                # Use remaining bytes for last block
                blocksize = size - tsize
            iolist.append({'offset':offset, 'write':iswrite, 'size':blocksize})
            offset += blocksize
            tsize += blocksize

        return iolist

    def _mem_alloc(self, size, aligned=False):
        """Allocate memory for use in C library functions"""
        dbuffer = None
        if aligned:
            # Allocate aligned buffer
            dbuffer = ctypes.c_void_p()
            self.libc.posix_memalign(ctypes.byref(dbuffer), self.PAGESIZE, size)
        else:
            # Allocate regular buffer
            dbuffer = self.libc.malloc(size)
        # Add allocated buffer so it can be freed
        self.fbuffers.append(dbuffer)
        return dbuffer

    def _getlock(self, name, fd, lock_type=None, offset=0, length=0, lock=None, tlock=False):
        """Get byte range lock on file given by file descriptor"""
        rn = self.random.randint(0,9999)
        stype = fcntl.F_SETLK
        if lock_type == fcntl.F_UNLCK:
            lstr = "UNLOCK"
            if not lock or rn >= 100*self.unlock:
                # Do not unlock file
                return
            self.nunlock += 1
        else:
            if tlock:
                # Just do TLOCK
                lstr = "TLOCK "
                stype = fcntl.F_GETLK
                if rn >= 100*self.tlock:
                    # No lock, so no tlock
                    return
                self.ntlock += 1
            else:
                lstr = "LOCK  "
                if rn >= 100*self.lock:
                    # No lock
                    return
                self.nlock += 1
            if lock_type is None:
                # Choose lock: read or write
                if self._percent(50):
                    lock_type = fcntl.F_RDLCK
                else:
                    lock_type = fcntl.F_WRLCK
            if not tlock:
                # LOCK is requested, but do TLOCK before actual lock
                self._getlock(name, fd, lock_type=lock_type, offset=offset, length=length, lock=lock, tlock=True)
        fstr = ""
        if offset == 0 and length == 0 and lstr == "LOCK  ":
            fstr = " full file"
        self._dprint("DBG4", "%s  %s %d @ %d (%s)%s" % (lstr, name, length, offset, LOCKMAP[lock_type], fstr))
        lockdata = struct.pack('hhllhh', lock_type, 0, offset, length, 0, 0)
        return fcntl.fcntl(fd, stype, lockdata)

    def _do_io(self, **kwargs):
        """Read or write to the given file descriptor"""
        fd       = kwargs.pop("fd", None)
        write    = kwargs.pop("write", False)
        offset   = kwargs.pop("offset", 0)
        size     = kwargs.pop("size", 0)
        fileobj  = kwargs.pop("fileobj", None)
        lockfull = kwargs.pop("lockfull", True)
        lockout  = None

        if self.iodelay > 0.0:
            time.sleep(self.iodelay)

        # Set file offset to read/write
        os.lseek(fd, offset, os.SEEK_SET)

        if write:
            if self.random and not lockfull:
                # Lock file segment
                lockout = self._getlock(fileobj.name, fd, lock_type=fcntl.F_WRLCK, offset=offset, length=size)
            data = 'x' * size
            self._dprint("DBG5", "WRITE   %s %d @ %d" % (fileobj.name, size, offset))

            if self.direct:
                # Direct I/O -- use native write function
                count = self.libc.write(fd, self.wbuffer, size)
            else:
                # Buffered I/O
                count = os.write(fd, data)
                if self._percent(self.fsync):
                    self._dprint("DBG4", "FSYNC   %s" % fileobj.name)
                    self.nfsync += 1
                    os.fsync(fd)

            self.nwrite += 1
            self.wbytes += count
            fsize = offset + count
            if fileobj.size < fsize:
                fileobj.size = fsize
        else:
            if self.random and not lockfull:
                # Lock file segment
                lockout = self._getlock(fileobj.name, fd, lock_type=fcntl.F_RDLCK, offset=offset, length=size)
            self._dprint("DBG5", "READ    %s %d @ %d" % (fileobj.name, size, offset))

            if self.direct:
                # Direct I/O -- use native read function
                count = self.libc.read(fd, self.rbuffer, size)
            else:
                # Buffered I/O
                data = os.read(fd, size)
                count = len(data)
            self.rbytes += count
            self.nread += 1

        if self.random and not lockfull:
            # Unlock file segment
            self._getlock(fileobj.name, fd, lock_type=fcntl.F_UNLCK, offset=offset, length=size, lock=lockout)
        return count

    def _do_file(self):
        """Operate on a file, create, read, truncate, etc."""
        self.absfile = ""
        # Number of files available
        nlen = len(self.n_files)
        if self.up_minfiles == 0 and nlen > self.minfiles:
            self.minfiles = self.bot_minfiles
            self.up_minfiles = 1
        if self.up_minfiles > 0 and nlen < self.minfiles:
            self.minfiles = self.top_minfiles
            self.up_minfiles = 0

        if nlen > self.minfiles and self._percent(self.trunc):
            # Truncate file using the file name
            fileobj = self._get_fileobj()
            self.absfile = os.path.join(self.datadir, fileobj.name)
            # Choose new size at random
            nsize = self.random.randint(0, fileobj.size + self.wsizedev)
            self._dprint("DBG2", "TRUNC   %s %d -> %d" % (fileobj.name, fileobj.size, nsize))
            out = self.libc.truncate(self.absfile, nsize)
            if out == -1:
                err = ctypes.get_errno()
                if hasattr(fileobj, 'srcname') and err == errno.ENOENT:
                    # Make sure not to fail if it is a broken symbolic link
                    self._dprint("DBG2", "TRUNC   %s: broken symbolic link" % fileobj.name)
                    return
                raise OSError(err, os.strerror(err), fileobj.name)
            else:
                self.ntrunc += 1
                fileobj.size = nsize
            return

        if nlen > self.minfiles and self._percent(self.rename):
            # Rename file
            fileobj = self._get_fileobj()
            name = self._newname()
            self.absfile = os.path.join(self.datadir, fileobj.name)
            newfile = os.path.join(self.datadir, name)
            self._dprint("DBG2", "RENAME  %s -> %s" % (fileobj.name, name))
            os.rename(self.absfile, newfile)
            self.nrename += 1
            fileobj.name = name
            return

        if nlen > self.minfiles and self._percent(self.remove):
            # Remove file
            fileobj = self._get_fileobj()
            self.absfile = os.path.join(self.datadir, fileobj.name)
            self._dprint("DBG2", "REMOVE  %s" % fileobj.name)
            os.unlink(self.absfile)
            self.nremove += 1
            self.n_files.pop(self.findex)
            return

        if nlen > self.minfiles and self._percent(self.link):
            # Create hard link
            name = self._newname()
            self.absfile = os.path.join(self.datadir, name)
            index = 0
            while True:
                index += 1
                fileobj = self._get_fileobj()
                if not hasattr(fileobj, 'srcname'):
                    # This file is not a symbolic link, use it
                    break
                if index >= 10:
                    self.absfile = os.path.join(self.datadir, fileobj.name)
                    raise Exception("Unable to find a valid source file for hard link")
            srcfile = os.path.join(self.datadir, fileobj.name)
            self._dprint("DBG2", "LINK    %s -> %s" % (name, fileobj.name))
            os.link(srcfile, self.absfile)
            self.nlink += 1
            linkobj = FileObj(name=name, size=fileobj.size)
            self.n_files.append(linkobj)
            return

        if nlen > self.minfiles and self._percent(self.slink):
            # Create symbolic link
            name = self._newname()
            self.absfile = os.path.join(self.datadir, name)
            index = 0
            while True:
                index += 1
                fileobj = self._get_fileobj()
                if not hasattr(fileobj, 'srcname'):
                    # This file is not a symbolic link, use it
                    break
                if index >= 10:
                    self.absfile = os.path.join(self.datadir, fileobj.name)
                    raise Exception("Unable to find a valid source file for symbolic link")
            self._dprint("DBG2", "SLINK   %s -> %s" % (name, fileobj.name))
            os.symlink(fileobj.name, self.absfile)
            self.nslink += 1
            slinkobj = FileObj(name=name, size=fileobj.size, srcname=fileobj.name)
            self.n_files.append(slinkobj)
            return

        if nlen > self.minfiles and self._percent(self.readdir):
            # Read directory
            count = self.random.randint(1,99)
            self._dprint("DBG2", "READDIR %s maxentries: %d" % (self.datadir, count))
            self.absfile = self.datadir
            fd = self.libc.opendir(self.datadir)
            index = 0
            while True:
                dirent = self.libc.readdir(fd)
                if dirent == 0 or index >= count:
                    break
                index += 1
            out = self.libc.closedir(fd)
            self.nreaddir += 1
            return

        # Select type of open: read, write or rdwr
        total = self.read + self.write
        rn = self.random.randint(0,9999)
        if rn < 100*self.read:
            oflags = os.O_RDONLY
            oflist = ["O_RDONLY"]
        elif rn < 100*total:
            oflags = os.O_WRONLY
            oflist = ["O_WRONLY"]
        else:
            oflags = os.O_RDWR
            oflist = ["O_RDWR"]

        # Set create file flag
        if nlen < self.minfiles:
            # Create at least self.minfiles before any other operation
            cflag = True
        else:
            cflag = self._percent(self.create)

        if cflag:
            # Create new name
            name = self._newname()
            fileobj = FileObj(name=name, size=0)
            self.n_files.append(fileobj)
            if oflags == os.O_RDONLY:
                # Creating file, must be able to write
                oflags = os.O_WRONLY
                oflist = ["O_WRONLY"]
            oflags |= os.O_CREAT
            oflist.append("O_CREAT")
        else:
            # Use name chosen at random
            fileobj = self._get_fileobj()

        if "O_RDONLY" not in oflist and self._percent(self.osync):
            # Add O_SYNC flag when opening file for writing
            oflags |= os.O_SYNC
            oflist.append("O_SYNC")
            self.nosync += 1

        if self.direct:
            # Open file for direct I/O
            oflags |= os.O_DIRECT
            oflist.append("O_DIRECT")

        # Select random or sequential I/O
        sstr = "sequen"
        if self._percent(self.randio):
            sstr = "random"

        ostr = "|".join(oflist)

        fd = None
        index = 0
        is_symlink = False
        while fd is None:
            try:
                index += 1
                if hasattr(fileobj, 'srcname'):
                    is_symlink = True
                self.absfile = os.path.join(self.datadir, fileobj.name)
                self._dprint("DBG2", "OPEN    %s %s %s" % (fileobj.name, sstr, ostr))
                fd = os.open(self.absfile, oflags)
                st = os.fstat(fd)
                if is_symlink:
                    self._dprint("DBG6", "OPEN    %s inode:%d symlink" % (fileobj.name, st.st_ino))
                    absfile = os.path.join(self.datadir, fileobj.srcname)
                    st = os.stat(absfile)
                    self._dprint("DBG6", "OPEN    %s inode:%d src:%s" % (fileobj.name, st.st_ino, fileobj.srcname))
                else:
                    self._dprint("DBG6", "OPEN    %s inode:%d" % (fileobj.name, st.st_ino))
            except OSError as openerr:
                if is_symlink and openerr.errno == errno.ENOENT:
                    self._dprint("DBG2", "OPEN    %s: broken symbolic link" % fileobj.name)
                    if index >= 10:
                        # Do not exit execution, just return to select another operation
                        return
                    # Choose a new name at random
                    fileobj = self._get_fileobj()
                    is_symlink = False
                else:
                    # Unknown error
                    raise
        self.nopen += 1

        # Get file size for writing
        size = int(abs(self.random.gauss(self.fsizeavg, self.fsizedev)))

        odgrade = False
        if oflags & os.O_WRONLY == os.O_WRONLY:
            lock_type = fcntl.F_WRLCK
            iolist = self._getiolist(size, True)
        elif oflags & os.O_RDWR == os.O_RDWR:
            lock_type = None
            iolist  = self._getiolist(size, True)
            iolist += self._getiolist(size, False)
            if self._percent(self.odgrade):
                odgrade = True
        else:
            lock_type = fcntl.F_RDLCK
            size = fileobj.size
            if size == 0:
                # File does not have any data, at least try to read one block
                size = self.rsize
            iolist = self._getiolist(size, False)

        if sstr == "random":
            # Shuffle I/O list for random access
            self.random.shuffle(iolist)

        # Lock full file if necessary
        lockfull = False
        if self._percent(self.lockfull):
            lockfull = True
            lockfout = self._getlock(fileobj.name, fd, lock_type=lock_type, offset=0, length=0)

        if nlen > self.minfiles and "O_RDONLY" not in oflist and self._percent(self.ftrunc):
            # Truncate file using the file descriptor
            # Choose new size at random
            nsize = self.random.randint(0, fileobj.size + self.wsizedev)
            self._dprint("DBG2", "FTRUNC  %s %d -> %d" % (fileobj.name, fileobj.size, nsize))
            os.ftruncate(fd, nsize)
            self.nftrunc += 1
            fileobj.size = nsize

        # Read or write the file
        for item in iolist:
            if self.runtime > 0 and time.time() >= self.s_time + self.runtime:
                # Runtime has been reached
                break
            self._do_io(**dict(fd=fd, fileobj=fileobj, lockfull=lockfull, **item))

        if lockfull:
            # Unlock full file
            self._getlock(fileobj.name, fd, lock_type=fcntl.F_UNLCK, offset=0, length=0, lock=lockfout)

        fdr = None
        fdroffset = 0
        if odgrade:
            # Need for open downgrade:
            # First, the file has been opened for read and write
            # Second, open file again for reading
            # Then close read and write file descriptor
            self._dprint("DBG2", "OPENDGR %s" % fileobj.name)
            fdr = os.open(self.absfile, os.O_RDONLY)
            self.nopendgr += 1
            count = self._do_io(fd=fdr, offset=fdroffset, size=self.rsize, fileobj=fileobj)
            fdroffset += count

        # Close main file descriptor
        self._dprint("DBG3", "CLOSE   %s" % fileobj.name)
        os.close(fd)
        self.nclose += 1

        if odgrade:
            for i in xrange(10):
                count = self._do_io(fd=fdr, offset=fdroffset, size=self.rsize, fileobj=fileobj)
                fdroffset += count
            self._dprint("DBG3", "CLOSE   %s" % fileobj.name)
            os.close(fdr)
            self.nclose += 1

        return

    def get_mountpoint(self):
        """Get mount point from data directory"""
        path = self.datadir
        st1 = os.stat(path)
        while path != os.sep:
            # Get parent directory
            parpath = os.path.realpath(os.path.join(path, os.pardir))
            st2 = os.stat(parpath)
            # Compare device ids from current and parent directories
            if st1.st_dev != st2.st_dev:
                break;
            path = parpath
        return path

    def run_process(self, tid=0):
        """Main loop for each process"""
        ret = 0
        stime = time.time()
        self.tid = tid
        self.n_index = 1
        self.n_files = []
        self.s_time  = stime

        # Setup signal handler to gracefully terminate process
        signal.signal(signal.SIGTERM, stop_handler)

        # Set file base name according to the number processes
        self.bidx = 1 + max(2, len("{0:x}".format(max(0,self.nprocs-1))))
        self.basename = "f{0:0{width}X}".format(self.tid, width=self.bidx-1)

        if self.createlogs:
            # Open a log file for each process
            if self.nprocs <= 10:
                self.logfile = self.logbase + "_%d.log" % self.tid
            elif self.nprocs <= 100:
                self.logfile = self.logbase + "_%02d.log" % self.tid
            elif self.nprocs <= 1000:
                self.logfile = self.logbase + "_%03d.log" % self.tid
            else:
                self.logfile = self.logbase + "_%04d.log" % self.tid
            self.open_log(self.logfile)

        # Read top level directory and populate file database when
        # a previous instance was ran on the same top level directory
        self._get_tree()

        # Create random object and initialized seed for process
        self.random = Random()
        self.random.seed(self.seed + tid)

        if self.direct:
            # Round up to nearest PAGESIZE boundary
            rsize = self.rsize + (self.PAGESIZE - self.rsize)%self.PAGESIZE
            wsize = self.wsize + (self.PAGESIZE - self.wsize)%self.PAGESIZE
            self._dprint("DBG7", "Allocating aligned read buffer of size %d" % rsize)
            self.rbuffer = self._mem_alloc(rsize, aligned=True)
            self._dprint("DBG7", "Allocating aligned write buffer of size %d" % wsize)
            self.wbuffer = self._mem_alloc(wsize, aligned=True)
            pdata = ctypes.create_string_buffer('x' * wsize)
            self.libc.memcpy(self.wbuffer, pdata, wsize);

        count = 0
        while True:
            try:
                self._do_file()
            except TermSignal:
                # SIGTERM has been raised, so stop running and send stats
                break
            except Exception:
                errstr = "ERROR on file object %s (process #%d)\n" % (self.absfile, self.tid)
                errstr += "Directory i-node: %d\n" % self.datadir_st.st_ino
                ioerror = traceback.format_exc()
                self._dprint("INFO", errstr+ioerror)
                ret = 1
                break
            ctime = time.time()
            if self.runtime > 0 and ctime >= stime + self.runtime:
                # Runtime has been reached
                break
            count += 1
        if self.queue:
            # Send all counts to main process
            self.queue.put(["RBYTES",   self.rbytes])
            self.queue.put(["WBYTES",   self.wbytes])
            self.queue.put(["NOPEN",    self.nopen])
            self.queue.put(["NOPENDGR", self.nopendgr])
            self.queue.put(["NOSYNC",   self.nosync])
            self.queue.put(["NCLOSE",   self.nclose])
            self.queue.put(["NREAD",    self.nread])
            self.queue.put(["NWRITE",   self.nwrite])
            self.queue.put(["NFSYNC",   self.nfsync])
            self.queue.put(["NRENAME",  self.nrename])
            self.queue.put(["NREMOVE",  self.nremove])
            self.queue.put(["NTRUNC",   self.ntrunc])
            self.queue.put(["NFTRUNC",  self.nftrunc])
            self.queue.put(["NLINK",    self.nlink])
            self.queue.put(["NSLINK",   self.nslink])
            self.queue.put(["NREADDIR", self.nreaddir])
            self.queue.put(["NLOCK",    self.nlock])
            self.queue.put(["NTLOCK",   self.ntlock])
            self.queue.put(["NUNLOCK",  self.nunlock])
            self.queue.put(["RETVALUE", ret])

        if self.direct:
            self._dprint("DBG7", "Free data buffers")
            for dbuffer in self.fbuffers:
                self.libc.free(dbuffer)
        self.close_log()
        return ret

    def run(self):
        """Main function where all processes are started"""
        errors = 0
        if self.seed is None:
            # Create random seed
            self.seed = int(1000.0*time.time())

        self.dprint("INFO", "System:  %s" % " ".join(os.uname()))
        self.dprint("INFO", "Command: %s" % " ".join(sys.argv))

        # Main seed so run can be reproduced
        self.dprint("INFO", "SEED = %d" % self.seed)
        # Flush log file descriptor to make sure above info is not written
        # to all log files when using multiple logs for each subprocess
        self.flush_log()
        stime = time.time()

        if not os.path.exists(self.datadir):
            # Create top level directory if it does not exist
            os.mkdir(self.datadir, 0777)
        self.datadir_st = os.stat(self.datadir)

        # Get mount stats for mount point
        mtpoint = self.get_mountpoint()
        cmd = "mountstats %s" % mtpoint
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        pstdout, pstderr = process.communicate()
        for line in pstdout.split("\n"):
            regex = re.search("Stats for\s+(.*):", line)
            if regex:
                self.dprint("INFO", regex.group(1))
            else:
                regex = re.search("NFS mount options:.*", line)
                if regex:
                    self.dprint("INFO", regex.group(0))

        if self.nprocs > 1:
            # setup interprocess queue
            self.queue = JoinableQueue()
            processes = []
            for i in xrange(self.nprocs):
                # Run each subprocess with its own process id (tid)
                # The process id is used to set the random number generator
                # and also to have each process work with different files
                process = Process(target=self.run_process, kwargs={'tid':self.tid})
                processes.append(process)
                process.start()
                self.tid += 1
            done = False
            while not done:
                # Wait for a short time so main process does not hog the CPU
                # by checking the queue continuously
                time.sleep(0.1)
                while not self.queue.empty():
                    # Get any pending messages from any of the processes
                    level, msg = self.queue.get()
                    # Check if message is a valid count first
                    if level == "RBYTES":
                        self.rbytes += msg
                    elif level == "WBYTES":
                        self.wbytes += msg
                    elif level == "NOPEN":
                        self.nopen += msg
                    elif level == "NOPENDGR":
                        self.nopendgr += msg
                    elif level == "NOSYNC":
                        self.nosync += msg
                    elif level == "NCLOSE":
                        self.nclose += msg
                    elif level == "NREAD":
                        self.nread += msg
                    elif level == "NWRITE":
                        self.nwrite += msg
                    elif level == "NFSYNC":
                        self.nfsync += msg
                    elif level == "NRENAME":
                        self.nrename += msg
                    elif level == "NREMOVE":
                        self.nremove += msg
                    elif level == "NTRUNC":
                        self.ntrunc += msg
                    elif level == "NFTRUNC":
                        self.nftrunc += msg
                    elif level == "NLINK":
                        self.nlink += msg
                    elif level == "NSLINK":
                        self.nslink += msg
                    elif level == "NREADDIR":
                        self.nreaddir += msg
                    elif level == "NLOCK":
                        self.nlock += msg
                    elif level == "NTLOCK":
                        self.ntlock += msg
                    elif level == "NUNLOCK":
                        self.nunlock += msg
                    elif level == "RETVALUE":
                        if msg != 0:
                            errors += 1
                            if self.exiterr:
                                # Exit on first error
                                for process in list(processes):
                                    process.terminate()
                                break
                    else:
                        # Message is not any of the valid counts,
                        # so treat it as a debug message
                        self.dprint(level, msg)
                # Check if any process has finished
                for process in list(processes):
                    if not process.is_alive():
                        process.join()
                        if not self.exiterr and abs(process.exitcode):
                            errors += 1
                        processes.remove(process)
                        if len(processes) == 0:
                            done = True
                            break
        else:
            # Only one process to run, just run the function
            out = self.run_process(tid=self.tid)
            if out != 0:
                errors += 1
        # Set seed to make sure if this function is called again a different
        # set of operations will be called
        self.seed += self.nprocs
        delta = time.time() - stime

        # Display stats
        formatstr.UNIT_SEP = " "
        readbytes  = str_units(self.rbytes)
        readbps    = str_units(self.rbytes/delta)
        writebytes = str_units(self.wbytes)
        writebps   = str_units(self.wbytes/delta)
        self.dprint("INFO", "==================STATS===================")
        self.dprint("INFO", "OPEN:    % 7d" % self.nopen)
        self.dprint("INFO", "OPENDGR: % 7d" % self.nopendgr)
        self.dprint("INFO", "CLOSE:   % 7d" % self.nclose)
        self.dprint("INFO", "OSYNC:   % 7d" % self.nosync)
        self.dprint("INFO", "READ:    % 7d, % 10s, % 10s/s" % (self.nread, readbytes, readbps))
        self.dprint("INFO", "WRITE:   % 7d, % 10s, % 10s/s" % (self.nwrite, writebytes, writebps))
        self.dprint("INFO", "FSYNC:   % 7d" % self.nfsync)
        self.dprint("INFO", "RENAME:  % 7d" % self.nrename)
        self.dprint("INFO", "REMOVE:  % 7d" % self.nremove)
        self.dprint("INFO", "TRUNC:   % 7d" % self.ntrunc)
        self.dprint("INFO", "FTRUNC:  % 7d" % self.nftrunc)
        self.dprint("INFO", "LINK:    % 7d" % self.nlink)
        self.dprint("INFO", "SLINK:   % 7d" % self.nslink)
        self.dprint("INFO", "READDIR: % 7d" % self.nreaddir)
        self.dprint("INFO", "LOCK:    % 7d" % self.nlock)
        self.dprint("INFO", "TLOCK:   % 7d" % self.ntlock)
        self.dprint("INFO", "UNLOCK:  % 7d" % self.nunlock)
        if errors > 0:
            self.dprint("INFO", "ERRORS:  % 7d" % errors)
        self.dprint("INFO", "TIME:    % 7d secs" % delta)
